
import os
import re
import shutil
import stat

# https://github.com/takluyver/pyxdg/blob/1d23e483ae869ee9532aca43b133cc43f63626a3/xdg/BaseDirectory.py
def get_runtime_dir(strict=True):
    try:
        return os.environ['XDG_RUNTIME_DIR']
    except KeyError:
        if strict:
            raise

        import getpass
        fallback = '/tmp/opensnitch-' + getpass.getuser()
        create = False

        try:
            # This must be a real directory, not a symlink, so attackers can't
            # point it elsewhere. So we use lstat to check it.
            st = os.lstat(fallback)
        except OSError as e:
            import errno
            if e.errno == errno.ENOENT:
                create = True
            else:
                raise
        else:
            # The fallback must be a directory
            if not stat.S_ISDIR(st.st_mode):
                os.unlink(fallback)
                create = True
            # Must be owned by the user and not accessible by anyone else
            elif (st.st_uid != os.getuid()) \
              or (st.st_mode & (stat.S_IRWXG | stat.S_IRWXO)):
                os.rmdir(fallback)
                create = True

        if create:
            os.mkdir(fallback, 0o700)

        return fallback

def get_run_opensnitch_dir():
    rdir = get_runtime_dir(False)
    if 'opensnitch' not in rdir:
        rdir = os.path.join(rdir, 'opensnitch')
        try:
            os.makedirs(rdir, 0o700)
        except:
            pass

    return rdir


class Autostart():
    def __init__(self):
        desktopFile = 'opensnitch_ui.desktop'
        self.systemDesktop = os.path.join('/usr/share/applications', desktopFile)
        self.systemAutostart = os.path.join('/etc/xdg/autostart', desktopFile)
        if not os.path.isfile(self.systemAutostart) and os.path.isfile('/usr' + self.systemAutostart):
            self.systemAutostart = '/usr' + self.systemAutostart
        self.userAutostart = os.path.join(xdg_config_home, 'autostart', desktopFile)

    def _copyfile(self, src, dst):
        """copy file (.desktop) to dst. Ignore exception if src and dst are equal"""
        try:
            shutil.copyfile(src, dst)
        except shutil.SameFileError:
            pass

    def createUserDir(self):
        if not os.path.isdir(xdg_config_home):
            os.makedirs(xdg_config_home, 0o700)
        if not os.path.isdir(os.path.dirname(self.userAutostart)):
            os.makedirs(os.path.dirname(self.userAutostart), 0o755)

    def isEnabled(self):
        ret = False
        if os.path.isfile(self.userAutostart):
             ret = True
             lines = open(self.userAutostart, 'r').readlines()
             for line in lines:
                 if re.search("^Hidden=true", line, re.IGNORECASE):
                     ret = False
                     break
        elif os.path.isfile(self.systemAutostart):
            ret = True
        return ret

    def enable(self, mode=True):
        self.createUserDir()
        if mode == True:
            if os.path.isfile(self.systemAutostart) and os.path.isfile(self.userAutostart):
                os.remove(self.userAutostart)
            elif os.path.isfile(self.systemDesktop):
                self._copyfile(self.systemDesktop, self.userAutostart)
        else:
            if os.path.isfile(self.systemAutostart):
                self._copyfile(self.systemAutostart, self.userAutostart)
                with open(self.userAutostart, 'a') as f:
                    f.write('Hidden=true\n')
            elif os.path.isfile(self.userAutostart):
                os.remove(self.userAutostart)

    def disable(self):
        self.enable(False)


_home = os.path.expanduser('~')
xdg_config_home = os.environ.get('XDG_CONFIG_HOME') or os.path.join(_home, '.config')
xdg_runtime_dir = get_runtime_dir(False)
xdg_current_desktop = os.environ.get('XDG_CURRENT_DESKTOP')
xdg_current_session = os.environ.get('XDG_SESSION_TYPE')

xdg_opensnitch_dir = get_run_opensnitch_dir()



from PyQt5 import Qt, QtCore
from PyQt5.QtWidgets import QApplication

# PyQt5 >= v5.15.8 (28/01/2023) (#821)
if hasattr(Qt, 'QItemDelegate'):
    from PyQt5.Qt import QItemDelegate, QStyleOptionViewItem
else:
    from PyQt5.QtWidgets import QItemDelegate, QStyleOptionViewItem


class ColorizedDelegate(QItemDelegate):
    TYPE = 'views'

    HMARGIN = 0
    VMARGIN = 1

    def __init__(self, parent=None, *args, actions={}):
        QItemDelegate.__init__(self, parent, *args)
        self._actions = actions
        self.modelColumns = parent.model().columnCount()
        self._style = QApplication.style()

    def setConfig(self, actions):
        self._actions = actions

    #@profile_each_line
    def paint(self, painter, option, index):
        """Override default widget style to personalize it with our own.
        """
        if self._actions.get('actions') == None:
            return super().paint(painter, option, index)
        if not index.isValid():
            return super().paint(painter, option, index)
        cellValue = index.data(QtCore.Qt.DisplayRole)
        if cellValue == None:
            return super().paint(painter, option, index)

        # initialize new QStyleOptionViewItem with the default options of this
        # cell.
        option = QStyleOptionViewItem(option)

        # by default use item's default attributes.
        # if we modify any of them, set it to False
        nocolor=True

        # don't call these functions in for-loops
        cellRect = QtCore.QRect(option.rect)
        curColumn = index.column()
        curRow = index.row()
        cellAlignment = option.displayAlignment
        defaultPen = painter.pen()
        defaultBrush = painter.brush()

        self._style = QApplication.style()
        # get default margins in order to respect them.
        # option.widget is the QTableView
        hmargin = self._style.pixelMetric(
            self._style.PM_FocusFrameHMargin, None, option.widget
        ) + 1
        vmargin = self._style.pixelMetric(
            self._style.PM_FocusFrameVMargin, None, option.widget
        ) + 1

        # set default margins for this cell
        cellRect.adjust(hmargin, vmargin, -painter.pen().width(), -painter.pen().width())

        for conf in self._actions['actions']:
            action = self._actions['actions'][conf]
            # if the corresponding plugin is not loaded, action will be dict
            # instead of instanceof(Plugin...)
            if type(action) is dict:
                continue

            if self.TYPE not in action.TYPE:
                print("colorizeddelegate: skipping action of type:", action.TYPE)
                continue
            modified = action.run(self,
                             (painter,
                              option,
                              index,
                              self._style,
                              self.modelColumns,
                              curRow,
                              curColumn,
                              defaultPen,
                              defaultBrush,
                              cellAlignment,
                              cellRect,
                              cellValue)
                             )
            if modified[0]:
                nocolor=False

        if nocolor:
            super().paint(painter, option, index)



from PyQt5.QtCore import QObject, QCoreApplication as QC
from google.protobuf import json_format

import opensnitch.proto as proto
ui_pb2, ui_pb2_grpc = proto.import_()

from opensnitch.nodes import Nodes
from .enums import *
from .rules import *
from .chains import *
from .exprs import *
from .profiles import *

class Firewall(QObject):
    __instance = None

    @staticmethod
    def instance():
        if Firewall.__instance == None:
            Firewall.__instance = Firewall()
        return Firewall.__instance

    def __init__(self, parent=None):
        QObject.__init__(self)
        self._nodes = Nodes.instance()
        self.rules = Rules(self._nodes)
        self.chains = Chains(self._nodes)

    def switch_rules(self, key, old_pos, new_pos):
        pass

    def add_rule(self, addr, rule):
        return self.rules.add(addr, rule)

    def insert_rule(self, addr, rule, position=0):
        return self.rules.insert(addr, rule, position)

    def update_rule(self, addr, uuid, rule):
        return self.rules.update(addr, uuid, rule)

    def delete_rule(self, addr, uuid):
        return self.rules.delete(addr, uuid)

    def change_rule_field(self, addr, uuid, field, value):
        addr, chain = self.get_rule_by_uuid(uuid)
        if chain is None:
            return None, None

        if field == Rules.FIELD_ENABLED:
            chain.Rules[0].Enabled = value
        elif field == Rules.FIELD_TARGET:
            chain.Rules[0].Target = value
        return self.update_rule(addr, uuid, chain)

    def enable_rule(self, addr, uuid, enable):
        addr, chain = self.get_rule_by_uuid(uuid)
        if chain is None:
            return None, None

        chain.Rules[0].Enabled = enable
        return self.update_rule(addr, uuid, chain)

    def filter_rules(self, nail):
        """
        """
        chains = []
        for addr in self._nodes.get_nodes():
            node = self._nodes.get_node(addr)
            if not 'firewall' in node:
                return chains
            for n in node['firewall'].SystemRules:
                for c in n.Chains:
                    for r in c.Rules:
                        add_rule = False
                        if nail == r.UUID:
                            add_rule = True

                        if nail in c.Family or \
                                nail in c.Hook or \
                                nail in r.Description or \
                                nail in r.Target or \
                                nail in r.TargetParameters:
                            add_rule = True
                        else:
                            for e in r.Expressions:
                                if add_rule:
                                    break
                                expr_vals = "".join("{0} {1}".format(h.Key, h.Value) for h in e.Statement.Values)
                                #print(nail in expr_vals, r.Description)
                                if nail in e.Statement.Op or \
                                        nail in e.Statement.Name or \
                                        nail in e.Statement.Values or \
                                        nail in expr_vals:
                                    add_rule = True

                        if add_rule:
                            chains.append(Rules.to_array(addr, c, r))

        return chains

    def filter_by_table(self, addr, table, family):
        """get rules by table"""
        chains = []
        node = self._nodes.get_node(addr)
        if not 'firewall' in node:
            return chains
        for n in node['firewall'].SystemRules:
            for c in n.Chains:
                for r in c.Rules:
                    if c.Table == table and c.Family == family:
                        chains.append(Rules.to_array(addr, c, r))

        return chains

    def filter_by_chain(self, addr, table, family, chain, hook):
        """get rules by chain"""
        chains = []
        node = self._nodes.get_node(addr)
        if not 'firewall' in node:
            return chains
        for n in node['firewall'].SystemRules:
            for c in n.Chains:
                for r in c.Rules:
                    if c.Table == table and c.Family == family and c.Name == chain and c.Hook == hook:
                        chains.append(Rules.to_array(addr, c, r))

        return chains

    def swap_rules(self, view, addr, uuid, old_pos, new_pos):
        return self.rules.swap(view, addr, uuid, old_pos, new_pos)

    def get_rule_by_uuid(self, uuid):
        """get rule by uuid, in string format
        """
        if uuid == "":
            return None, None
        for addr in self._nodes.get_nodes():
            node = self._nodes.get_node(addr)
            if not 'fwrules' in node:
                continue
            r = node['fwrules'].get(uuid)
            if r != None:
                return addr, r

        return None, None

    def get_protorule_by_uuid(self, addr, uuid):
        """get protobuffer rule by uuid.
        """
        return self.rules.get_by_uuid(addr, uuid)

    def get_node_rules(self, addr):
        return self.rules.get_by_node(addr)

    def get_chains(self):
        return self.chains.get()

    def get_rules(self):
        return self.rules.get()

    def rule_to_json(self, rule):
        return Rules.to_json(rule)

    def apply_profile(self, node_addr, json_profile):
        """
        Apply a profile to the firewall configuration.

        Given a chain (table+family+type+hook), apply its policy, and any rules
        defined.
        """
        try:
            holder = ui_pb2.FwChain()
            profile = json_format.Parse(json_profile, holder)

            fwcfg = self._nodes.get_node(node_addr)['firewall']
            for sdx, n in enumerate(fwcfg.SystemRules):
                for cdx, c in enumerate(n.Chains):

                    if c.Hook.lower() == profile.Hook and \
                            c.Type.lower() == profile.Type and \
                            c.Family.lower() == profile.Family and \
                            c.Table.lower() == profile.Table:

                        fwcfg.SystemRules[sdx].Chains[cdx].Policy = profile.Policy
                        for r in profile.Rules:
                            temp_c = ui_pb2.FwChain()
                            temp_c.CopyFrom(c)
                            del temp_c.Rules[:]
                            temp_c.Rules.extend([r])

                            if self.rules.is_duplicated(node_addr, temp_c):
                                continue
                            self.add_rule(node_addr, temp_c)

                        self.rules.rulesUpdated.emit()
                        return True, ""
        except Exception as e:
            return False, "{0}".format(e)

        return False, QC.translate("firewall", "profile not applied")

    def delete_profile(self, node_addr, json_profile):
        try:
            holder = ui_pb2.FwChain()
            profile = json_format.Parse(json_profile, holder)

            fwcfg = self._nodes.get_node(node_addr)['firewall']
            for sdx, n in enumerate(fwcfg.SystemRules):
                for cdx, c in enumerate(n.Chains):
                    if c.Hook.lower() == profile.Hook and \
                            c.Type.lower() == profile.Type and \
                            c.Family.lower() == profile.Family and \
                            c.Table.lower() == profile.Table:

                        if profile.Policy == ProfileDropInput.value:
                            profile.Policy = ProfileAcceptInput.value

                        del_candidates = []
                        for rdx, r in enumerate(c.Rules):
                            for pr in profile.Rules:
                                if r.UUID == pr.UUID:
                                    # we cannot delete the rule here, otherwise
                                    # we'd modify the items of the loop.
                                    del_candidates.append(rdx)
                        if len(del_candidates) > 0:
                            for rdx in del_candidates:
                                if rdx == len(c.Rules): # last rule
                                    rdx = rdx - 1
                                self.delete_rule(node_addr, c.Rules[rdx].UUID)

        except Exception as e:
            return False, "{0}".format(e)
        return False, QC.translate("firewall", "profile not deleted")




# common configuration to Highlight Action column
commonDelegateConfig = {
    "name": "commonDelegateConfig",
    "created": "",
    "updated": "",
    "actions": {
        "highlight": {
            "enabled": True,
            "cells": [
                {
                    "text": ["allow", "\u2713 online"],
                    "operator": "==",
                    "cols": [1, 2, 3],
                    "color": "green",
                    "bgcolor": "",
                    "alignment": ["center"]
                },
                {
                    "text": ["deny", "\u2613 offline"],
                    "cols": [1, 2, 3],
                    "color": "red",
                    "bgcolor": "",
                    "alignment": ["center"]
                },
                {
                    "text": ["reject"],
                    "cols": [1, 2, 3],
                    "color": "purple",
                    "bgcolor": "",
                    "alignment": ["center"]
                }
            ],
            "rows": []
        }
    }
}

# firewall rules configuration to Highlight Enabled and Action columns
fwDelegateConfig = {
    "name": "defaultFWDelegateConfig",
    "created": "",
    "updated": "",
    "actions": {
        "highlight": {
            "enabled": True,
            "cells": [
                {
                    "text": [
                        "allow",
                        "True",
                        "accept",
                        "jump",
                        "masquerade",
                        "snat",
                        "dnat",
                        "tproxy",
                        "queue",
                        "redirect",
                        "True",
                        "ACCEPT"

                    ],
                    "cols": [7, 10],
                    "color": "green",
                    "bgcolor": "",
                    "alignment": ["center"]
                },
                {
                    "text": [
                        "deny",
                        "False",
                        "drop",
                        "DROP",
                        "stop"
                    ],
                    "cols": [7, 10],
                    "color": "red",
                    "bgcolor": "",
                    "alignment": ["center"]
                },
                {
                    "text": [
                        "reject",
                        "return"
                    ],
                    "cols": [7, 10],
                    "color": "purple",
                    "bgcolor": "",
                    "alignment": ["center"]
                }

            ],
            "rows": []
    }
  }
}

# rules configuration to Highlight Enabled and Action columns
rulesDelegateConfig = {
    "name": "defaultRulesDelegateConfig",
    "created": "",
    "updated": "",
    "actions": {
        "highlight": {
            "enabled": True,
            "cells": [
                {
                    "text": ["allow", "True"],
                    "cols": [3, 4],
                    "color": "green",
                    "bgcolor": "",
                    "alignment": ["center"]
                    },
                    {
                    "text": ["deny", "False"],
                    "cols": [3, 4],
                    "color": "red",
                    "bgcolor": "",
                    "alignment": ["center"]
                    },
                    {
                    "text": ["reject"],
                    "cols": [3, 4],
                    "color": "purple",
                    "bgcolor": "",
                    "alignment": ["center"]
                    }
            ],
            "rows": []
        }
    }
}

netstatDelegateConfig = {
    "name": "netstatDelegateConfig",
    "created": "",
    "updated": "",
    "actions": {
        "highlight": {
            "enabled": True,
            "cells": [
                {
                    "text": ["LISTEN"],
                    "cols": [2],
                    "color": "green",
                    "bgcolor": "",
                    "alignment": ["center"]
                },
                {
                    "text": ["CLOSE"],
                    "cols": [2],
                    "color": "red",
                    "bgcolor": "",
                    "alignment": ["center"]
                },
                {
                    "text": ["Established"],
                    "cols": [2],
                    "color": "blue",
                    "bgcolor": "",
                    "alignment": ["center"]
                },
                {
                    "text": [
                        "TCP_SYN_SENT", "TCP_SYN_RECV",
                        "TCP_FIN_WAIT1", "TCP_FIN_WAIT2",
                        "TCP_TIME_WAIT", "TCP_CLOSE_WAIT",
                        "TCP_LAST_ACK", "TCP_CLOSING",
                        "TCP_NEW_SYNC_RECV"
                    ],
                    "cols": [2],
                    "color": "",
                    "bgcolor": "",
                    "alignment": ["center"]
                }
            ],
            "rows": []
        }
    }
}



# Generated by the gRPC Python protocol compiler plugin. DO NOT EDIT!
import grpc

from . import ui_pb2 as ui__pb2


class UIStub(object):
  # missing associated documentation comment in .proto file
  pass

  def __init__(self, channel):
    """Constructor.

    Args:
      channel: A grpc.Channel.
    """
    self.Ping = channel.unary_unary(
        '/protocol.UI/Ping',
        request_serializer=ui__pb2.PingRequest.SerializeToString,
        response_deserializer=ui__pb2.PingReply.FromString,
        )
    self.AskRule = channel.unary_unary(
        '/protocol.UI/AskRule',
        request_serializer=ui__pb2.Connection.SerializeToString,
        response_deserializer=ui__pb2.Rule.FromString,
        )
    self.Subscribe = channel.unary_unary(
        '/protocol.UI/Subscribe',
        request_serializer=ui__pb2.ClientConfig.SerializeToString,
        response_deserializer=ui__pb2.ClientConfig.FromString,
        )
    self.Notifications = channel.stream_stream(
        '/protocol.UI/Notifications',
        request_serializer=ui__pb2.NotificationReply.SerializeToString,
        response_deserializer=ui__pb2.Notification.FromString,
        )
    self.PostAlert = channel.unary_unary(
        '/protocol.UI/PostAlert',
        request_serializer=ui__pb2.Alert.SerializeToString,
        response_deserializer=ui__pb2.MsgResponse.FromString,
        )


class UIServicer(object):
  # missing associated documentation comment in .proto file
  pass

  def Ping(self, request, context):
    # missing associated documentation comment in .proto file
    pass
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def AskRule(self, request, context):
    # missing associated documentation comment in .proto file
    pass
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def Subscribe(self, request, context):
    # missing associated documentation comment in .proto file
    pass
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def Notifications(self, request_iterator, context):
    # missing associated documentation comment in .proto file
    pass
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def PostAlert(self, request, context):
    # missing associated documentation comment in .proto file
    pass
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')


def add_UIServicer_to_server(servicer, server):
  rpc_method_handlers = {
      'Ping': grpc.unary_unary_rpc_method_handler(
          servicer.Ping,
          request_deserializer=ui__pb2.PingRequest.FromString,
          response_serializer=ui__pb2.PingReply.SerializeToString,
      ),
      'AskRule': grpc.unary_unary_rpc_method_handler(
          servicer.AskRule,
          request_deserializer=ui__pb2.Connection.FromString,
          response_serializer=ui__pb2.Rule.SerializeToString,
      ),
      'Subscribe': grpc.unary_unary_rpc_method_handler(
          servicer.Subscribe,
          request_deserializer=ui__pb2.ClientConfig.FromString,
          response_serializer=ui__pb2.ClientConfig.SerializeToString,
      ),
      'Notifications': grpc.stream_stream_rpc_method_handler(
          servicer.Notifications,
          request_deserializer=ui__pb2.NotificationReply.FromString,
          response_serializer=ui__pb2.Notification.SerializeToString,
      ),
      'PostAlert': grpc.unary_unary_rpc_method_handler(
          servicer.PostAlert,
          request_deserializer=ui__pb2.Alert.FromString,
          response_serializer=ui__pb2.MsgResponse.SerializeToString,
      ),
  }
  generic_handler = grpc.method_handlers_generic_handler(
      'protocol.UI', rpc_method_handlers)
  server.add_generic_rpc_handlers((generic_handler,))




class ConnFields():
    Time = 0
    Node = 1
    Action = 2
    Protocol = 3
    SrcIP = 4
    SrcPort = 5
    DstIP = 6
    DstHost = 7
    DstPort = 8
    UID = 9
    PID = 10
    Process = 11
    Cmdline = 12
    CWD = 13
    Rule = 14

class RuleFields():
    """These fields must be in the order defined in the DB"""
    Time = 0
    Node = 1
    Name = 2
    Enabled = 3
    Precedence = 4
    Action = 5
    Duration = 6
    OpType = 7
    OpSensitive = 8
    OpOperand = 9
    OpData = 10
    Description = 11
    NoLog = 12
    Created = 13

class AlertFields():
    Time = 0
    Node = 1
    Type = 2
    Action = 3
    Priority = 4
    What = 5
    Body = 6
    Status = 7



from PyQt5 import Qt, QtCore
from PyQt5.QtGui import QColor, QPalette

from opensnitch.plugins import PluginBase, PluginSignal

# PyQt5 >= v5.15.8 (#821)
if hasattr(Qt, 'QStyle'):
    from PyQt5.Qt import QStyle
else:
    from PyQt5.QtWidgets import QStyle


class Highlight(PluginBase):
    """Customizes QTablewView cells via QItemDelegates.
    Format:
    "highlight": {
      "cells": [
        {
          "text": ["allow", "True"],
          "cols": [3, 4],
          "color": "green",
          "bgcolor": "",
          "alignment": ["center"]
        }
      ]
      "rows":[
        {
          "text": ["False"],
          "cols": [3],
          "color": "black",
          "bgcolor": "darkgray"
        }
      ]
    }

    cells: rules will be applied only on individual cells.
    rows: rules will be applied to rows on the given columns.

    Fields:
      text: will match any of the given texts (the comparison is an OR operation).
      cols: look for patterns on these columns.
      color: colorizes the color of the text.
      bgcolor: colorizes the background color of the cell.
    etc.

    Color names: https://doc.qt.io/qt-6/qcolor.html#predefined-colors

    Notes:
     - There're 3 default configurations that are applied on the views:
         - commonDelegateConfig, defaultRulesDelegateConfig and
         defaultFWDelegateConfig

        Creating/Copying these configurations under
        XDG_CONFIG_HOME/.config/opensnitch/actions/ allows to overwrite and hence
        personalize the views highlighting colors.

     - The order of the "rules" are applied from top to bottom, meaning that
     the last "rule" of the "cells" or "rows" arrays will override previous
     rules if there're conflicts. For example:

     [Rules tab]
        Action | Name                | Enabled | ...
         deny  | allow-always-telnet |  False  | ...

    Config:
       "rows": [
          {
              "text": ["False"],
              "cols": [3],
              "color": "black",
              "bgcolor": "darkgray"
          },
          {
              "text": ["allow"],
              "cols": [4],
              "color": "white",
              "bgcolor": "green"
          }
       ]

    In this example, the background color of this row will always be green,
    because the latest "rule" will be the one that will be applied.
    It may have more sense to put the first "rule" last, to properly colorize
    not enabled rules.

    """
    name = "Highlight"
    version = "0.1"
    author = "opensnitch"
    created = ""
    modified = ""
    #enabled = True

    # where this plugin is allowed
    TYPE = [PluginBase.TYPE_VIEWS]

    MARGINS = "margins"
    ALIGNMENT = "alignment"
    # QtCore.Qt.AlignCenter
    ALIGN_CENTER = "center"
    # QtCore.Qt.AlignHCenter
    ALIGN_HCENTER = "hcenter"
    # QtCore.Qt.AlignVCenter
    ALIGN_VCENTER = "vcenter"

    COLOR = "color"
    BGCOLOR = "bgcolor"
    FONT = "font"
    CELLS = "cells"
    ROWS = "rows"
    COLS = "cols"
    TEXT = "text"


    def __init__(self, config=None):
        # original json config received
        self._config = config
        self._last_visited_row = -1
        self._rowcells = ""
        self.signal_in.connect(self.cb_signal)

    def get_name(self):
        return self.name

    def is_enabled(self):
        return self.enabled

    def set_enabled(self, enable):
        self.enabled = enable

    def get_description(self):
        return self.description

    def configure(self):
        # TODO: allow to configure rules from the GUI
        #  - add button(s) to the rules editor dialog, to allow choose a color
        #  for a row based on patterns.
        #  - create a dialog to configure new rules.
        pass

    def _compile(self, what):
        cell_list = self._config.get(what)
        if cell_list == None:
            print("highlight plugin: configuration has no '{0}' configuration".format(what))
            return
        for cell in cell_list:
            for item in cell:
                # colors
                if (item == Highlight.COLOR or item == Highlight.BGCOLOR):
                    if cell[item] != "" and cell[item] is not None:
                        try:
                            cell[item] = QColor(cell[item])
                        except Exception as e:
                            cell[item] = None
                    else:
                        cell[item] = None

                # alignments
                if item == Highlight.ALIGNMENT:
                    cell[item] = self.getAlignment(cell[item])

                # TODO: fonts
                #if item == Highlight.FONT:
                #    cell[item] = self.getFont(cell[item])

        return self._config

    def compile(self):
        """transform json items to Qt objects.
        These items are transformed:
            - color (to QColor), bgcolor (to QColor), alignment (to Qt.Align*),
            font (to QFont TODO)

            Return the original json object transformed.
        """
        self._config = self._compile(Highlight.CELLS)
        self._config = self._compile(Highlight.ROWS)
        return self._config

    def run(self, parent, args):
        """Highlight cells or rows based on patterns.

        Return if the cell was modified.

        Keyword arguments:
            args -- tuple of options.
        """
        #parent == ColorizedDelegate
        painter = args[0]
        option = args[1]
        index = args[2]
        style = args[3]
        modelColumns = args[4]
        curRow = args[5]
        curColumn = args[6]
        defaultPen = args[7]
        defaultBrush = args[8]
        cellAlignment = args[9]
        cellRect = args[10]
        cellValue = args[11]
        #print(type(self), ">", type(parent), ">", type(option.widget))

        # signal that this cell has been modified
        modified = False

        cells = self._config.get(Highlight.CELLS)
        rows = self._config.get(Highlight.ROWS)

        if cells:
            for cell in cells:
                if curColumn not in cell[Highlight.COLS]:
                    continue
                if cellValue not in cell[Highlight.TEXT]:
                    continue
# TODO
#                if cell['operator'] == 'simple' and cellValue != cell['text']:
#                    continue
#                elif cell['text'] not in cellValue:
#                    continue

                cellColor = cell.get(Highlight.COLOR)
                cellBgColor = cell.get(Highlight.BGCOLOR)
                if cell.get(Highlight.ALIGNMENT) != None:
                    cellAlignment = cell[Highlight.ALIGNMENT]
                if cell.get(Highlight.MARGINS) != None:
                    cellRect.adjust(
                        int(cell[Highlight.MARGINS][self.HMARGIN]),
                        int(cell[Highlight.MARGINS][self.VMARGIN]),
                        -defaultPen.width(),
                        -defaultPen.width()
                    )

                modified=True
                self.paintCell(
                    style,
                    painter,
                    option,
                    index,
                    defaultPen,
                    defaultBrush,
                    cellAlignment,
                    cellRect,
                    cellColor,
                    cellBgColor,
                    cellValue)

        if len(rows) == 0:
            return (modified,)

        # get row's cells only for the first cell of the row,
        # then reuse them for the rest of the cells of the current row.
        if curRow != self._last_visited_row:
            try:
                # TODO: check for NoneType
                self._rowcells = " ".join(
                    [index.sibling(curRow, col).data() for col in range(0, modelColumns)]
                )
            except:
                pass
        self._last_visited_row = curRow

        for row in rows:
            skip = True
            for text in row[Highlight.TEXT]:
                if text in self._rowcells:
                    skip = False
            if skip:
                continue

            cellColor = row.get(Highlight.COLOR)
            cellBgColor = row.get(Highlight.BGCOLOR)
            if row.get(Highlight.ALIGNMENT) != None:
                cellAlignment = row[Highlight.ALIGNMENT]
            if row.get(Highlight.MARGINS) != None:
                cellRect.adjust(
                    int(row[Highlight.MARGINS][self.HMARGIN]),
                    int(row[Highlight.MARGINS][self.VMARGIN]),
                    -defaultPen.width(),
                    -defaultPen.width()
                )

            modified=True
            self.paintCell(
                style,
                painter,
                option,
                index,
                defaultPen,
                defaultBrush,
                cellAlignment,
                cellRect,
                cellColor,
                cellBgColor,
                cellValue)

        return (modified,)

    def paintCell(self,
                  style,
                  painter,
                  option,
                  index,
                  defaultPen,
                  defaultBrush,
                  cellAlignment,
                  cellRect,
                  cellColor,
                  cellBgColor,
                  cellValue):
        cellSelected = option.state & QStyle.State_Selected

        painter.save()
        # don't customize selected state
        if not cellSelected:
            if cellBgColor:
                painter.fillRect(option.rect, cellBgColor)

            if cellColor:
                defaultPen.setColor(cellColor)
        else:
            painter.fillRect(option.rect, option.palette.color(QPalette.Highlight))
            defaultPen.setColor(option.palette.color(QPalette.HighlightedText))

        painter.setPen(defaultPen)

        # setting option.displayAlignment has no effect here, so we need to
        # draw the text.
        # FIXME: Drawing the text though, the background color of the SelectedState is
        # altered.
        # If we called super().paint(), modifying option.palette.* would be
        # enough to change the text color, but it wouldn't be aligned:
        # option.palette.setColor(QPalette.Text, cellColor)
        style.drawItemText(painter, cellRect, cellAlignment, option.palette, True, cellValue)
        painter.restore()

    def getAlignment(self, alignments):
        alignFlags = 0
        for align in alignments:
            if align == Highlight.ALIGN_CENTER:
                alignFlags |= QtCore.Qt.AlignCenter
            elif align == Highlight.ALIGN_HCENTER:
                alignFlags |= QtCore.Qt.AlignHCenter
            elif align == Highlight.ALIGN_VCENTER:
                alignFlags |= QtCore.Qt.AlignVCenter

        if alignFlags == 0:
            return None

        return alignFlags

    def getFont(self, font):
        # TODO
        pass

    def stop(self):
        pass

    def cb_signal(self, signal):
        #print("Plugin.signal received:", self.name, signal)
        try:
            if signal['signal'] == PluginSignal.ENABLE:
                self.enabled = True
        except Exception as e:
            print("Plugin.Highlight.cb_signal() exception:", e)



from PyQt5 import QtCore
from PyQt5.QtGui import QRegion
from PyQt5.QtWidgets import QItemDelegate, QAbstractItemView, QPushButton, QWidget, QVBoxLayout
from PyQt5.QtCore import pyqtSignal

class UpDownButtonDelegate(QItemDelegate):
    clicked = pyqtSignal(int, QtCore.QModelIndex)

    UP=-1
    DOWN=1

    def paint(self, painter, option, index):
        if (
            isinstance(self.parent(), QAbstractItemView)
            and self.parent().model() is index.model()
        ):
            self.parent().openPersistentEditor(index)

    def createEditor(self, parent, option, index):
        w = QWidget(parent)
        w.setContentsMargins(0, 0, 0, 0)
        w.setAutoFillBackground(True)

        layout = QVBoxLayout(w)
        layout.setContentsMargins(0, 0, 0, 0)

        btnUp = QPushButton(parent)
        btnUp.setText("⇡")
        btnUp.setFlat(True)
        btnUp.clicked.connect(lambda: self._cb_button_clicked(self.UP, index))

        btnDown = QPushButton(parent)
        btnDown.setText("⇣")
        btnDown.setFlat(True)
        btnDown.clicked.connect(lambda: self._cb_button_clicked(self.DOWN, index))

        layout.addWidget(btnUp)
        layout.addWidget(btnDown)
        return w

    def _cb_button_clicked(self, action, idx):
        self.clicked.emit(action, idx)

    def updateEditorGeometry(self, editor, option, index):
        rect = QtCore.QRect(option.rect)
        minWidth = editor.minimumSizeHint().width()
        if rect.width() < minWidth:
            rect.setWidth(minWidth)
        editor.setGeometry(rect)
        # create a new mask based on the option rectangle, then apply it
        mask = QRegion(0, 0, option.rect.width(), option.rect.height())
        editor.setProperty('offMask', mask)
        editor.setMask(mask)



import os.path
import sys, glob
from PyQt5 import QtCore

from opensnitch.config import Config

# WA for #1373
qtm_home = "{0}/.qt_material/theme/".format(QtCore.QDir.homePath())
if qtm_home not in QtCore.QDir.searchPaths("icon"):
    QtCore.QDir.addSearchPath("icon", qtm_home)

class Themes():
    """Change GUI's appearance using qt-material lib.
    https://github.com/dunderlab/qt-material
    """
    THEMES_PATH = [
        os.path.expanduser("~/.config/opensnitch/"),
        os.path.dirname(sys.modules[__name__].__file__) + "/../.."
    ]
    __instance = None

    AVAILABLE = False
    IS_DARK = False
    try:
        from qt_material import apply_stylesheet as qtmaterial_apply_stylesheet
        from qt_material import list_themes as qtmaterial_themes
        AVAILABLE = True
    except Exception:
        print("Themes not available. Install qt-material if you want to change GUI's appearance: pip3 install qt-material.")

    @staticmethod
    def instance():
        if Themes.__instance == None:
            Themes.__instance = Themes()
        return Themes.__instance

    def __init__(self):
        self._cfg = Config.get()
        theme = self._cfg.getInt(self._cfg.DEFAULT_THEME, 0)

    def available(self):
        return Themes.AVAILABLE

    def get_saved_theme(self):
        theme = self._cfg.getSettings(self._cfg.DEFAULT_THEME)
        theme_density = self._cfg.getSettings(self._cfg.DEFAULT_THEME_DENSITY_SCALE)
        if theme_density == "" or theme_density == None:
            theme_density = '0'

        if not Themes.AVAILABLE:
            return 0, "", theme_density

        try:
            if theme != "" and theme != None:
                themes_list = self.list_themes()
                for idx, tml in enumerate(themes_list):
                    if theme in tml:
                        # 0 == System
                        return idx+1, theme, theme_density
        except Exception as e:
            print("Themes.get_saved_theme() error:", e)
        return 0, "", theme_density

    def save_theme(self, theme_idx, theme, density_scale):
        if not Themes.AVAILABLE:
            return

        self._cfg.setSettings(self._cfg.DEFAULT_THEME_DENSITY_SCALE, density_scale)
        if theme_idx == 0:
            self._cfg.setSettings(self._cfg.DEFAULT_THEME, "")
        else:
            self._cfg.setSettings(self._cfg.DEFAULT_THEME, theme)

    def load_theme(self, app):
        if not Themes.AVAILABLE:
            return

        try:
            theme_idx, theme_name, theme_density = self.get_saved_theme()
            if theme_name != "":
                invert = "light" in theme_name
                fname = os.path.basename(theme_name)
                Themes.IS_DARK = fname.startswith("dark")

                print("Using theme:", theme_idx, theme_name, "inverted:", invert, "dark:", Themes.IS_DARK)
                # TODO: load {theme}.xml.extra and .xml.css for further
                # customizations.
                extra_opts = {
                    'density_scale': theme_density
                }
                Themes.qtmaterial_apply_stylesheet(app, theme=theme_name,  invert_secondary=invert, extra=extra_opts)
        except Exception as e:
            print("Themes.load_theme() exception:", e)

    def change_theme(self, window, theme_name, extra={}):
        try:
            invert = "light" in theme_name
            fname = os.path.basename(theme_name)
            Themes.IS_DARK = fname.startswith("dark")

            Themes.qtmaterial_apply_stylesheet(window, theme=theme_name,  invert_secondary=invert, extra=extra)
        except Exception as e:
            print("Themes.change_theme() exception:", e, " - ", window, theme_name)

    def list_local_themes(self):
        themes = []
        if not Themes.AVAILABLE:
            return themes

        try:
            for tdir in self.THEMES_PATH:
                if not os.path.isdir(tdir):
                    continue
                themes += glob.glob(tdir + "/themes/*.xml")
        except Exception:
            pass
        finally:
            return themes

    def list_themes(self):
        themes = self.list_local_themes()
        if not Themes.AVAILABLE:
            return themes

        themes += Themes.qtmaterial_themes()
        return themes



from threading import Lock
import configparser
import threading
import glob
import os
import re
import locale

is_pyinotify_available = True
try:
    import pyinotify
except Exception as e:
    is_pyinotify_available = False
    print("Error importing pyinotify:", e)

DESKTOP_PATHS = tuple([
    os.path.join(d, 'applications')
    for d in os.getenv('XDG_DATA_DIRS', '/usr/share/').split(':')
])

class LinuxDesktopParser(threading.Thread):
    def __init__(self):
        threading.Thread.__init__(self)
        self.lock = Lock()
        self.daemon = True
        self.running = False
        self.apps = {}
        self.apps_by_name = {}
        self.get_locale()
        # some things are just weird
        # (not really, i don't want to keep track of parent pids
        # just because of icons though, this hack is way easier)
        self.fixes = {
            '/opt/google/chrome/chrome': '/opt/google/chrome/google-chrome',
            '/usr/lib/firefox/firefox': '/usr/lib/firefox/firefox.sh',
            '/usr/bin/pidgin.orig': '/usr/bin/pidgin'
        }

        self.terminal_icon = os.path.join(
           os.path.abspath(os.path.dirname(__file__)),
            "res/terminal.svg"
        )

        for desktop_path in DESKTOP_PATHS:
            if not os.path.exists(desktop_path):
                continue
            for desktop_file in glob.glob(os.path.join(desktop_path, '*.desktop')):
                self._parse_desktop_file(desktop_file)

        if is_pyinotify_available:
            self.start()

    def get_locale(self):
        try:
            self.locale = locale.getlocale()[0]
            self.locale_country = self.locale.split("_")[0]
        except Exception:
            self.locale = ""
            self.locale_country = ""

    def _parse_exec(self, cmd):
        try:
            is_flatpak = re.search(r'^/usr/[s]*bin/flatpak.*--command=([a-zA-Z0-9-_\/\.\+]+)', cmd)
            if is_flatpak:
                return is_flatpak.group(1)

            # remove stuff like %U
            cmd = re.sub( r'%[a-zA-Z]+', '', cmd)
            # remove 'env .... command'
            cmd = re.sub( r'^env\s+[^\s]+\s', '', cmd)
            # split && trim
            cmd = cmd.split(' ')[0].strip()
            # remove quotes
            cmd = re.sub( r'["\']+', '', cmd)

            # check if we need to resolve the path
            if len(cmd) > 0 and cmd[0] != '/':
                for path in os.environ["PATH"].split(os.pathsep):
                    filename = os.path.join(path, cmd)
                    if os.path.exists(filename):
                        cmd = filename
                        break

        except Exception as e:
            print("desktop_parser._parse_exec() exception:", e)

        return cmd

    def get_app_description(self, parser):
        try:
            desc = parser.get('Desktop Entry', 'Comment[%s]' % self.locale_country, raw=True, fallback=None)
            if desc == None:
                desc = parser.get('Desktop Entry', 'Comment[%s]' % self.locale, raw=True, fallback=None)

            if desc == None:
                desc = parser.get('Desktop Entry', 'Comment', raw=True, fallback=None)

            return desc
        except:
            return None

    @staticmethod
    def discover_app_icon(app_name):
        # more hacks
        # normally qt will find icons if the system is configured properly.
        # if it's not, qt won't be able to find the icon by using QIcon().fromTheme(""),
        # so we fallback to try to determine if the icon exist in some well known system paths.
        for idir in glob.glob('/usr/share/icons/*/*/apps/*'):
            if app_name in idir:
                return idir

        return None

    def _parse_desktop_file(self, desktop_path):
        parser = configparser.ConfigParser(strict=False)  # Allow duplicate config entries
        try:
            basename = os.path.basename(desktop_path)[:-8]
            parser.read(desktop_path, 'utf8')

            cmdline = parser.get('Desktop Entry', 'exec', raw=True, fallback=None)
            if cmdline == None:
                cmdline = parser.get('Desktop Entry', 'Exec', raw=True, fallback=None)
            if cmdline is None:
                return

            cmd  = self._parse_exec(cmdline)
            icon = parser.get('Desktop Entry', 'Icon', raw=True, fallback=None)
            name = parser.get('Desktop Entry', 'Name', raw=True, fallback=None)
            desc = self.get_app_description(parser)

            if name == "flatpak":
                return

            if icon == None:
                # Some .desktop files doesn't have the Icon entry
                # FIXME: even if we return an icon, if the DE is not properly configured,
                # it won't be loaded/displayed.
                icon = LinuxDesktopParser.discover_app_icon(basename)
                if icon == None:
                    icon = self.terminal_icon

            with self.lock:
                # The Exec entry may have an absolute path to a binary or just the binary with parameters.
                # /path/binary or binary, so save both
                self.apps[cmd] = (name, icon, desc, desktop_path)
                self.apps[basename] = (name, icon, desc, desktop_path)
                # if the command is a symlink, add the real binary too
                if os.path.islink(cmd):
                    link_to = os.path.realpath(cmd)
                    self.apps[link_to] = (name, icon, desc, desktop_path)
        except:
            print("Exception parsing .desktop file ", desktop_path)

    def get_info_by_path(self, path, default_icon):
        def_name = os.path.basename(path)
        # apply fixes
        for orig, to in self.fixes.items():
            if path == orig:
                path = to
                break

        app_name = self.apps.get(path)
        if app_name != None:
            return self.apps.get(path, (def_name, default_icon, "", None))

        app_name = self.apps.get(def_name)
        if app_name != None:
            return self.apps.get(def_name, (def_name, default_icon, "", None))

        # last try to get a default terminal icon
        icon = self.discover_app_icon(def_name)
        if icon == None:
            icon = self.terminal_icon

        return self.apps.get(def_name, (def_name, icon, "", None))

    def get_info_by_binname(self, name, default_icon):
        def_name = os.path.basename(name)
        return self.apps.get(def_name, (def_name, default_icon, None))

    def run(self):
        self.running = True
        wm = pyinotify.WatchManager()
        notifier = pyinotify.Notifier(wm)

        def inotify_callback(event):
            if event.mask == pyinotify.IN_CLOSE_WRITE:
                self._parse_desktop_file(event.pathname)

            elif event.mask == pyinotify.IN_DELETE:
                with self.lock:
                    for cmd, data in self.apps.items():
                        if data[2] == event.pathname:
                            del self.apps[cmd]
                            break

        for p in DESKTOP_PATHS:
            if os.path.exists(p):
                wm.add_watch(p,
                             pyinotify.IN_CLOSE_WRITE | pyinotify.IN_DELETE,
                             inotify_callback)
        notifier.loop()



from PyQt5 import QtCore
from PyQt5.QtGui import QColor, QStandardItemModel, QStandardItem
from PyQt5.QtSql import QSqlQueryModel, QSqlQuery, QSql
from PyQt5.QtWidgets import QTableView
from PyQt5.QtCore import QItemSelectionModel, pyqtSignal, QEvent
import time
import math

from PyQt5.QtCore import QCoreApplication as QC
class ColorizedQSqlQueryModel(QSqlQueryModel):
    """
        model=CustomQSqlQueryModel(
            modelData=
                {
                'colorize':
                      {'offline': (QColor(QtCore.Qt.red), 2)},
                'alignment': { Qt.AlignLeft, 2 }
                }
            )
    """
    RED   = QColor(QtCore.Qt.red)
    GREEN = QColor(QtCore.Qt.green)

    def __init__(self, modelData={}):
        QSqlQueryModel.__init__(self)
        self._model_data = modelData

    def data(self, index, role=QtCore.Qt.DisplayRole):
        if not index.isValid():
            return QSqlQueryModel.data(self, index, role)

        column = index.column()
        row = index.row()

        if role == QtCore.Qt.TextAlignmentRole:
            return QtCore.Qt.AlignCenter
        if role == QtCore.Qt.TextColorRole:
            for _, what in enumerate(self._model_data):
                d = QSqlQueryModel.data(self, self.index(row, self._model_data[what][1]), QtCore.Qt.DisplayRole)
                if column == self._model_data[what][1] and what in d:
                    return self._model_data[what][0]

        return QSqlQueryModel.data(self, index, role)

class ConnectionsTableModel(QStandardItemModel):
    rowCountChanged = pyqtSignal()

    #max rowid in the db; starts with 1, not with 0
    maxRowId = 0
    #previous total number of rows in the db when the filter was applied
    prevFiltRowCount = 0
    #total number of rows in the db when the filter was not applied
    prevNormRowCount = 0
    #total row count which must de displayed in the view
    totalRowCount = 0
    #new rows which must be added to the top of the rows displayed in the view
    prependedRowCount = 0

    db = None
    #original query string before we modify it
    origQueryStr = QSqlQuery()
    #modified query object
    realQuery = QSqlQuery()
    #previous original query string; used to check if the query has changed
    prevQueryStr = ''
    #whether or not the original query has a filter (a WHERE condition)
    isQueryFilter = False
    limit = None

    #a map for fast lookup or rows when filter is enabled
    #contains ranges of rowids and count of filter hits
    #range format {'from': <rowid>, 'to': <rowid>, 'hits':<int>}
    #including the 'from' rowid up to but NOT including the 'to' rowid
    map = []
    rangeSize = 1000
    #all unique/distinct values for each column will be stored here
    distinct = {'time':[], 'process':[], 'dst_host':[], 'dst_ip':[], 'dst_port':[], 'rule':[], 'node':[], 'protocol':[]}
    #what was the last rowid\time when the distinct value were updates
    distinctLastRowId = 0
    distinctLastUpdateTime = time.time()

    def __init__(self):
        self.headerLabels = [
            QC.translate("stats", "Time", "This is a word, without spaces and symbols.").replace(" ", ""),
            QC.translate("stats", "Node", "This is a word, without spaces and symbols.").replace(" ", ""),
            QC.translate("stats", "Action", "This is a word, without spaces and symbols.").replace(" ", ""),
            QC.translate("stats", "Destination", "This is a word, without spaces and symbols.").replace(" ", ""),
            QC.translate("stats", "Protocol", "This is a word, without spaces and symbols.").replace(" ", ""),
            QC.translate("stats", "Process", "This is a word, without spaces and symbols.").replace(" ", ""),
            QC.translate("stats", "Rule", "This is a word, without spaces and symbols.").replace(" ", ""),
        ]
        QStandardItemModel.__init__(self, 0, len(self.headerLabels))
        self.setHorizontalHeaderLabels(self.headerLabels)

    #Some QSqlQueryModel methods must be mimiced so that this class can serve as a drop-in replacement
    #mimic QSqlQueryModel.query()
    def query(self):
        return self

    #mimic QSqlQueryModel.query().lastQuery()
    def lastQuery(self):
        return self.origQueryStr

    #mimic QSqlQueryModel.query().lastError()
    def lastError(self):
        return self.realQuery.lastError()

    #mimic QSqlQueryModel.clear()
    def clear(self):
        pass

    def setQuery(self, q, db):
        self.origQueryStr = q
        self.db = db
        maxRowIdQuery = QSqlQuery(db)
        maxRowIdQuery.setForwardOnly(True)
        maxRowIdQuery.exec("SELECT MAX(rowid) FROM connections")
        maxRowIdQuery.first()
        value = maxRowIdQuery.value(0)
        self.maxRowId = 0 if value == '' else int(value)
        self.updateDistinctIfNeeded()
        self.limit = int(q.split(' ')[-1]) if q.split(' ')[-2] == 'LIMIT' else None
        self.isQueryFilter = True if ("LIKE '%" in q and "LIKE '% %'" not in q) or 'Action = "' in q else False

        self.realQuery = QSqlQuery(db)
        isTotalRowCountChanged = False
        isQueryChanged = False
        if self.prevQueryStr != q:
            isQueryChanged = True
        if self.isQueryFilter:
            if isQueryChanged:
                self.buildMap()
            largestRowIdInMap = self.map[0]['from']
            newRowsCount = self.maxRowId - largestRowIdInMap
            self.prependedRowCount = 0

            if newRowsCount > 0:
                starttime = time.time()
                self.realQuery.setForwardOnly(True)
                for offset in range(0, newRowsCount, self.rangeSize):
                    lowerBound = largestRowIdInMap + offset
                    upperBound = min(lowerBound + self.rangeSize, self.maxRowId)
                    part1, part2 = q.split('ORDER')
                    qStr = part1 + 'AND rowid>'+ str(lowerBound) + ' AND rowid<=' + str(upperBound) + ' ORDER' + part2
                    self.realQuery.exec(qStr)
                    self.realQuery.last()
                    rowsInRange = max(0, self.realQuery.at()+1)
                    if self.map[0]['from'] - self.map[0]['to'] < self.rangeSize:
                        #consolidate with the previous range; we don't want many small ranges
                        self.map[0]['from'] = upperBound
                        self.map[0]['hits'] += rowsInRange
                    else:
                        self.map.insert(0, {'from':upperBound, 'to':lowerBound, 'hits':rowsInRange})
                    self.prependedRowCount += rowsInRange
                    if time.time() - starttime > 0.5:
                        #dont freeze the UI when fetching too many recent rows
                        break

            self.totalRowCount = 0
            for i in self.map:
                self.totalRowCount += i['hits']
            if self.totalRowCount != self.prevFiltRowCount:
                isTotalRowCountChanged = True
                self.prevFiltRowCount = self.totalRowCount
        else: #self.isQueryFilter == False
            self.prependedRowCount = self.maxRowId - self.prevNormRowCount
            self.totalRowCount = self.maxRowId
            if self.totalRowCount != self.prevNormRowCount:
                isTotalRowCountChanged = True
                self.prevNormRowCount = self.totalRowCount

        self.prevQueryStr = self.origQueryStr
        if isTotalRowCountChanged or self.prependedRowCount > 0 or isQueryChanged:
            self.rowCountChanged.emit()

    #fill self.map with data
    def buildMap(self):
        self.map = []
        q = QSqlQuery(self.db)
        q.setForwardOnly(True)
        self.updateDistinctIfNeeded(True)
        filterStr = self.getFilterStr()
        actionStr = self.getActionStr()
        #we only want to know the count of matching rows
        qStr = "SELECT COUNT(*) from connections WHERE (rowid> :lowerBound AND rowid<= :upperBound)"
        if actionStr:
            qStr += ' AND ' + actionStr
        matchStr = self.getMatch(filterStr) if filterStr else None
        if matchStr:
            qStr += ' AND ' + matchStr
        qStr += ' LIMIT ' + str(self.limit) if self.limit else ''
        q.prepare(qStr)

        totalRows = 0
        for offset in range(self.maxRowId, -1, -self.rangeSize):
            upperBound = offset
            lowerBound = max(0, upperBound - self.rangeSize)
            if (not filterStr and actionStr) or (filterStr and matchStr):
                #either 1) only action was present or 2) filter which has a match (with or without action)
                q.bindValue(":lowerBound", str(lowerBound))
                q.bindValue(":upperBound", str(upperBound))
                q.exec_()
                q.first()
                rowsInRange = int(q.value(0))
            else:
                rowsInRange = 0
            totalRows += rowsInRange
            self.map.append({'from':upperBound, 'to':lowerBound, 'hits':rowsInRange})
            if self.limit and totalRows >= self.limit:
                break

    #periodically keep track of all distinct values for each column
    #this is needed in order to build efficient queries when the filter is applied
    def updateDistinctIfNeeded(self, force=False):
        if (not force and (time.time() - self.distinctLastUpdateTime) < 10) or self.maxRowId == self.distinctLastRowId:
            return
        if (self.maxRowId < self.distinctLastRowId):
            #the db has been cleared, re-init the values
            self.distinctLastRowId = 0
            self.distinct = {'time':[], 'process':[], 'dst_host':[], 'dst_ip':[], 'dst_port':[], 'rule':[], 'node':[], 'protocol':[]}
        q = QSqlQuery(self.db)
        q.setForwardOnly(True)
        for column in self.distinct.keys():
            q.exec('SELECT DISTINCT ' + column + ' FROM connections WHERE rowid>'
            + str(self.distinctLastRowId) + ' AND rowid<=' + str(self.maxRowId))
            while q.next():
                if q.value(0) not in self.distinct[column]:
                    self.distinct[column].append(q.value(0))
        self.distinctLastRowId =self.maxRowId
        self.distinctLastUpdateTime = time.time()

    #refresh the viewport with data from the db
    #"value" is vertical scrollbar's value
    def refreshViewport(self, value, maxRowsInViewport):
        q = QSqlQuery(self.db)
        #sequential number of topmost/bottommost rows in viewport (numbering starts from the bottom with 1 not with 0)
        botRowNo = max(1, self.totalRowCount - (value + maxRowsInViewport-1))
        topRowNo = min(botRowNo + maxRowsInViewport-1, self.totalRowCount)

        if not self.isQueryFilter:
            part1, part2 = self.origQueryStr.split('ORDER')
            qStr = part1 + 'WHERE rowid>='+ str(botRowNo) + ' AND rowid<=' + str(topRowNo) + ' ORDER' + part2
        else:
            self.updateDistinctIfNeeded(True)
            #replace query part between WHERE and ORDER
            qStr = self.origQueryStr.split('WHERE')[0] + ' WHERE '
            actionStr = self.getActionStr()
            if actionStr:
                qStr += actionStr + " AND "
            #find inside the map the range(s) in which top and bottom rows are located
            total, offsetInRange, botRowFound, topRowFound = 0, None, False, False
            ranges = [{'from':0, 'to':0, 'hits':0}]
            for i in reversed(self.map):
                if total + i['hits'] >= botRowNo:
                    botRowFound = True
                if total + i['hits'] >= topRowNo:
                    topRowFound = True
                if botRowFound and i['hits'] > 0:
                    if i['to'] == ranges[-1]['from']:
                        #merge two adjacent ranges
                        ranges[-1]['from'] = i['from']
                        ranges[-1]['hits'] += i['hits']
                    else:
                        ranges.append(i.copy())
                if topRowFound:
                    offsetInRange = i['hits'] - (topRowNo - total)
                    break
                total += i['hits']

            rangeStr = ''
            if len(ranges) > 0:
                rangeStr = '('
                for r in ranges:
                    rangeStr += '(rowid>' + str(r['to']) + ' AND rowid<=' + str(r['from']) + ') OR '
                rangeStr = rangeStr[:-3] #remove trailing 'OR '
                rangeStr += ') AND '
            qStr += rangeStr

            filterStr = self.getFilterStr()
            matchStr = self.getMatch(filterStr) if filterStr else None
            if matchStr:
                qStr += matchStr + " AND "
            qStr = qStr[:-4] #remove trailing ' AND'
            qStr += ' ORDER '+ self.origQueryStr.split('ORDER')[1]

        q.exec(qStr)
        q.last()
        rowsFound = max(0, q.at()+1)
        if not self.isQueryFilter:
            q.seek(QSql.BeforeFirstRow)
        else:
            #position the db cursor on topRowNo
            q.seek(QSql.BeforeFirstRow if offsetInRange == 0 else offsetInRange-1)
        upperBound = min(maxRowsInViewport, rowsFound)
        self.setRowCount(upperBound)
        #only visible rows will be filled with data
        if upperBound > 0:
            #don't trigger setItem's signals for each cell, instead emit dataChanged for all cells
            self.blockSignals(True)
            for x in range(0, upperBound):
                q.next()
                for col in range(0, len(self.headerLabels)):
                    self.setItem(x, col, QStandardItem(q.value(col)))
            self.blockSignals(False)
            self.dataChanged.emit(self.createIndex(0,0), self.createIndex(upperBound, len(self.headerLabels)))

    #form a condition string for the query: if filterStr is (partially) present in any of the columns
    def getMatch (self, filterStr):
        match = {}
        for column in self.distinct.keys():
            match[column] = []
            for value in self.distinct[column]:
                if filterStr in value:
                    match[column].append(value)
        matchStr = None
        if any([match[col] for col in match]):
            matchStr = '( '
            if match['time']:
                matchStr += "time IN ('" + "','".join(match['time']) + "') OR"
            if match['process']:
                matchStr += "process IN ('" + "','".join(match['process']) + "') OR"
            if match['dst_host']:
                matchStr += " (dst_host != '' AND dst_host IN ('" + "','".join(match['dst_host']) + "') ) OR"
            if match['dst_ip']:
                matchStr += " (dst_host = '' AND dst_ip IN ('" + "','".join(match['dst_ip']) + "') ) OR"
            if match['dst_port']:
                matchStr += " dst_port IN ('" + "','".join(match['dst_port']) + "') OR"
            if match['rule']:
                matchStr += " rule IN ('" + "','".join(match['rule']) + "') OR"
            if match['node']:
                matchStr += " node IN ('" + "','".join(match['node']) + "') OR"
            if match['protocol']:
                matchStr += " protocol IN ('" + "','".join(match['protocol']) + "') OR"
            matchStr = matchStr[:-2] #remove trailing 'OR'
            matchStr += ' )'
        return matchStr

    #extract the filter string if any
    def getFilterStr(self):
        filterStr = None
        if "LIKE '%" in self.origQueryStr:
            filterStr = self.origQueryStr.split("LIKE '%")[1].split("%")[0]
        return filterStr

    #extract the action string if any
    def getActionStr(self):
        actionStr = None
        if 'WHERE Action = "' in self.origQueryStr:
            actionCond = self.origQueryStr.split('WHERE Action = "')[1].split('"')[0]
            actionStr = "action = '"+actionCond+"'"
        return actionStr

    def dumpRows(self):
        rows = []
        q = QSqlQuery(self.db)
        q.exec(self.origQueryStr)
        q.seek(QSql.BeforeFirstRow)
        while True:
            q.next()
            if q.at() == QSql.AfterLastRow:
                break
            row = []
            for col in range(0, len(self.headerLabels)):
                row.append(q.value(col))
            rows.append(row)
        return rows

class ConnectionsTableView(QTableView):
    # how many rows can potentially be displayed in viewport
    # the actual number of rows currently displayed may be less than this
    maxRowsInViewport = 0
    #vertical scroll bar
    vScrollBar = None

    def __init__(self, parent):
        QTableView.__init__(self, parent)
        #eventFilter to catch key up/down events and wheel events
        self.installEventFilter(self)
        self.verticalHeader().setVisible(False)
        self.horizontalHeader().setDefaultAlignment(QtCore.Qt.AlignCenter)
        self.horizontalHeader().setStretchLastSection(True)
        #the built-in vertical scrollBar of this view is always off
        self.setVerticalScrollBarPolicy(QtCore.Qt.ScrollBarAlwaysOff)

    def setVerticalScrollBar(self, vScrollBar):
        self.vScrollBar = vScrollBar
        self.vScrollBar.valueChanged.connect(self.onValueChanged)
        self.vScrollBar.setVisible(False)

    def setModel(self, model):
        super().setModel(model)
        model.rowCountChanged.connect(self.onRowCountChanged)
        model.rowsInserted.connect(self.onRowsInsertedOrRemoved)
        model.rowsRemoved.connect(self.onRowsInsertedOrRemoved)
        self.horizontalHeader().sortIndicatorChanged.disconnect()
        self.setSortingEnabled(False)

    #model().rowCount() is always <= self.maxRowsInViewport
    #stretch the bottom row; we don't want partial-height rows at the bottom
    #this will only trigger if rowCount value was changed
    def onRowsInsertedOrRemoved(self, parent, start, end):
        if self.model().rowCount() == self.maxRowsInViewport:
            self.verticalHeader().setStretchLastSection(True)
        else:
            self.verticalHeader().setStretchLastSection(False)

    def resizeEvent(self, event):
        super().resizeEvent(event)
        #refresh the viewport data based on new geometry
        self.calculateRowsInViewport()
        self.model().setRowCount(min(self.maxRowsInViewport, self.model().totalRowCount))
        self.model().refreshViewport(self.vScrollBar.value(), self.maxRowsInViewport)

    def calculateRowsInViewport(self):
        rowHeight = self.verticalHeader().defaultSectionSize()
        #we don't want partial-height rows in viewport, hence .floor()
        self.maxRowsInViewport = math.floor(self.viewport().height() / rowHeight)

    def onValueChanged(self, vSBNewValue):
        savedIndex = self.selectionModel().currentIndex()
        self.model().refreshViewport(vSBNewValue, self.maxRowsInViewport)
        #restore selection which was removed by model's refreshing the data
        self.selectionModel().setCurrentIndex(savedIndex, QItemSelectionModel.Rows | QItemSelectionModel.SelectCurrent)

    # if ( scrollbar at the top or row limit set):
    #   let new rows "push down" older rows without changing the scrollbar position
    # else:
    #   don't update data in viewport, only change scrollbar position.
    def onRowCountChanged(self):
        totalCount = self.model().totalRowCount
        scrollBar = self.vScrollBar
        scrollBar.setVisible(True if totalCount > self.maxRowsInViewport else False)
        scrollBarValue = scrollBar.value()
        if self.model().limit:
            newValue = min(scrollBarValue, self.model().limit - self.maxRowsInViewport)
            scrollBar.setMinimum(0)
            scrollBar.setMaximum( min(totalCount, self.model().limit) - self.maxRowsInViewport)
            if scrollBarValue != newValue:
                #setValue does not trigger valueChanged if new value is the same as old
                scrollBar.setValue(newValue)
            else:
                scrollBar.valueChanged.emit(newValue)
        else:
            scrollBar.setMinimum(0)
            scrollBar.setMaximum(max(0, totalCount - self.maxRowsInViewport))
            if scrollBarValue == 0:
                scrollBar.valueChanged.emit(0)
            elif scrollBarValue > 0:
                if self.model().prependedRowCount == 0:
                    scrollBar.valueChanged.emit(scrollBarValue)
                else:
                    scrollBar.setValue(scrollBarValue + self.model().prependedRowCount)

    def onKeyUp(self):
        if self.selectionModel().currentIndex().row() == 0:
            self.vScrollBar.setValue(self.vScrollBar.value() - 1)

    def onKeyDown(self):
        if self.selectionModel().currentIndex().row() == self.maxRowsInViewport - 1:
            self.vScrollBar.setValue(self.vScrollBar.value() + 1)

    def onKeyHome(self):
        self.vScrollBar.setValue(0)
        self.selectionModel().setCurrentIndex(self.model().createIndex(0, 0), QItemSelectionModel.Rows | QItemSelectionModel.SelectCurrent)

    def onKeyEnd(self):
        self.vScrollBar.setValue(self.vScrollBar.maximum())
        self.selectionModel().setCurrentIndex(self.model().createIndex(min(self.maxRowsInViewport, self.model().totalRowCount) - 1, 0), QItemSelectionModel.Rows | QItemSelectionModel.SelectCurrent)

    def onKeyPageUp(self):
        #scroll up only when on the first row
        if self.selectionModel().currentIndex().row() != 0:
            return
        self.vScrollBar.setValue(self.vScrollBar.value() - self.maxRowsInViewport)

    def onKeyPageDown(self):
        #scroll down only when on the last row
        if self.selectionModel().currentIndex().row() != self.maxRowsInViewport - 1:
            return
        self.vScrollBar.setValue(self.vScrollBar.value() + self.maxRowsInViewport)

    def eventFilter(self, obj, event):
        if event.type() == QEvent.KeyPress:
            if event.key() == QtCore.Qt.Key_Up:
                self.onKeyUp()
            elif event.key() == QtCore.Qt.Key_Down:
                self.onKeyDown()
            elif event.key() == QtCore.Qt.Key_Home:
                self.onKeyHome()
            elif event.key() == QtCore.Qt.Key_End:
                self.onKeyEnd()
            elif event.key() == QtCore.Qt.Key_PageUp:
                self.onKeyPageUp()
            elif event.key() == QtCore.Qt.Key_PageDown:
                self.onKeyPageDown()
        elif event.type() == QEvent.Wheel:
            self.vScrollBar.wheelEvent(event)
        return False



#   Copyright (C) 2018      Simone Margaritelli
#                 2019-2025 Gustavo Iñiguez Goia
#
#   This file is part of OpenSnitch.
#
#   OpenSnitch is free software: you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   OpenSnitch is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with OpenSnitch.  If not, see <http://www.gnu.org/licenses/>.

from packaging.version import Version
import importlib
from opensnitch.utils import Versions

# Protobuffers compiled with protobuf < 3.20.0 are incompatible with
# protobuf >= 4.0.0
# https://github.com/evilsocket/opensnitch/wiki/GUI-known-problems#gui-does-not-show-up
#
# In order to solve this issue, we provide several protobuffers:
# proto.ui_pb2* for protobuf >= 4.0.0
# proto.pre3200.ui_pb2* for protobuf >= 3.6.0 and < 3.20.0
#
# To avoid import errors, each protobuffer must be placed in its own directory,
# and the name of the protobuffer files must be named with the syntax
# <prefix>_pb2.py/<prefix>_pb2_grpc.py:
#  ui_pb2.py and ui_pb2_grpc.py

default_pb = "opensnitch.proto.ui_pb2"
default_grpc = "opensnitch.proto.ui_pb2_grpc"
old_pb = "opensnitch.proto.pre3200.ui_pb2"
old_grpc = "opensnitch.proto.pre3200.ui_pb2_grpc"

def import_():
    """load the protobuffer needed based on the grpc and protobuffer version
    installed in the system.
    """
    try:
        gui_version, grpc_version, proto_version = Versions.get()
        proto_ver = default_pb
        grpc_ver = default_grpc

        if Version(proto_version) < Version("3.20.0"):
            proto_ver = old_pb
            grpc_ver = old_grpc

        return importlib.import_module(proto_ver), importlib.import_module(grpc_ver)
    except Exception as e:
        print("error importing protobuffer: ", repr(e))
        return importlib.import_module(default_pb, default_grpc)




from PyQt5 import QtCore, QtGui

class RestrictChars(QtGui.QValidator):
    result = QtCore.pyqtSignal(object)

    def __init__(self, restricted_chars, *args, **kwargs):
        QtGui.QValidator.__init__(self, *args, **kwargs)
        self._restricted_chars = restricted_chars

    def validate(self, value, pos):
        # allow to delete all characters
        if len(value) == 0:
            return QtGui.QValidator.Intermediate, value, pos

        # user can type characters or paste them.
        # pos value when pasting can be any number, depending on where did the
        # user paste the characters.
        for char in self._restricted_chars:
            if char in value:
                self.result.emit(QtGui.QValidator.Invalid)
                return QtGui.QValidator.Invalid, value, pos

        self.result.emit(QtGui.QValidator.Acceptable)
        return QtGui.QValidator.Acceptable, value, pos




import grpc

Simple = "simple"
TLSSimple = "tls-simple"
TLSMutual = "tls-mutual"

NO_CLIENT_CERT = "no-client-cert"
REQ_CERT = "req-cert"
REQ_ANY_CERT = "req-any-cert"
VERIFY_CERT = "verify-cert"
REQ_AND_VERIFY_CERT = "req-and-verify-cert"


def load_file(file_path):
    try:
        with open(file_path, "rb") as f:
            return f.read()
    except Exception as e:
        print("auth: error loading {0}: {1}".format(file_path, e))

    return None


def get_tls_credentials(ca_cert, server_cert, server_key):
    """return a new gRPC credentials object given a server cert and key file.
    https://grpc.io/docs/guides/auth/#python
    """
    try:
        cacert = load_file(ca_cert)
        cert = load_file(server_cert)
        cert_key = load_file(server_key)
        auth_nodes = False if cacert == None else True

        return grpc.ssl_server_credentials(
            ((cert_key, cert),),
            cacert,
            auth_nodes
        )
    except Exception as e:
        print("get_tls_credentials error:", e)
        return None



from slugify import slugify
import os

from PyQt5.QtCore import QCoreApplication as QC

from opensnitch.config import Config
from opensnitch.dialogs.prompt import _constants

def truncate_text(text, max_size=64):
    if len(text) > max_size:
        text = text[:max_size] + "..."
    return text

def set_elide_text(widget, text, max_size=64):
    text = truncate_text(text, max_size)
    widget.setText(text)

def get_rule_name(rule, is_list):
    rule_temp_name = slugify("%s %s" % (rule.action, rule.duration))
    if is_list:
        rule_temp_name = "%s-list" % rule_temp_name
    else:
        rule_temp_name = "%s-simple" % rule_temp_name
    rule_temp_name = slugify("%s %s" % (rule_temp_name, rule.operator.data))

    return rule_temp_name[:128]

def get_popup_message(is_local, node, app_name, con):
    """
    _get_popup_message helps constructing the message that is displayed on
    the pop-up dialog. Example:
        curl is connecting to www.opensnitch.io on TCP port 443
    """
    app_name = truncate_text(app_name)

    message = "<b>{0}</b>".format(app_name)
    if not is_local:
        message = QC.translate("popups", "<b>Remote</b> process {0} running on <b>{1}</b>".format(
            message,
            node.split(':')[1])
        )

    msg_action = QC.translate("popups", "is connecting to <b>{0}</b> on {1} port {2}".format(
        con.dst_host or con.dst_ip,
        con.protocol.upper(),
        con.dst_port )
    )

    # icmp port is 0 (i.e.: no port)
    if con.dst_port == 0:
        msg_action = QC.translate("popups", "is connecting to <b>{0}</b>, {1}".format(
            con.dst_host or con.dst_ip,
            con.protocol.upper() )
        )

    if con.dst_port == 53 and con.dst_ip != con.dst_host and con.dst_host != "":
        msg_action = QC.translate("popups", "is attempting to resolve <b>{0}</b> via {1}, {2} port {3}".format(
            con.dst_host,
            con.dst_ip,
            con.protocol.upper(),
            con.dst_port)
        )

    return "{0} {1}".format(message, msg_action)

def set_app_path(appPathLabel, app_name, app_args, con):
    # show the binary path if it's not part of the cmdline args:
    # cmdline: telnet 1.1.1.1 (path: /usr/bin/telnet.netkit)
    # cmdline: /usr/bin/telnet.netkit 1.1.1.1 (the binary path is part of the cmdline args, no need to display it)
    if con.process_path != "" and len(con.process_args) >= 1 and con.process_path not in con.process_args:
        appPathLabel.setToolTip("Process path: {0}".format(con.process_path))
        if app_name.lower() == app_args:
            set_elide_text(appPathLabel, "%s" % con.process_path)
        else:
            set_elide_text(appPathLabel, "(%s)" % con.process_path)
        appPathLabel.setVisible(True)
    elif con.process_path != "" and len(con.process_args) == 0:
        set_elide_text(appPathLabel, "%s" % con.process_path)
        appPathLabel.setVisible(True)
    else:
        appPathLabel.setVisible(False)
        appPathLabel.setText("")

    appPathLabel.setText(
        "".join(
            filter(str.isprintable, appPathLabel.text())
        )
    )

def set_app_args(argsLabel, app_name, app_args):
    # if the app name and the args are the same, there's no need to display
    # the args label (amule for example)
    if app_name.lower() != app_args:
        argsLabel.setVisible(True)
        set_elide_text(argsLabel, app_args, 256)
        argsLabel.setToolTip(app_args)
    else:
        argsLabel.setVisible(False)
        argsLabel.setText("")

    argsLabel.setText(
        "".join(
            filter(str.isprintable, argsLabel.text())
        )
    )

def set_app_description(appDescriptionLabel, description):
    if description != None and description != "":
        appDescriptionLabel.setVisible(True)
        appDescriptionLabel.setFixedHeight(50)
        appDescriptionLabel.setToolTip(description)
        set_elide_text(appDescriptionLabel, "%s" % description)
    else:
        appDescriptionLabel.setVisible(False)
        appDescriptionLabel.setFixedHeight(0)
        appDescriptionLabel.setText("")

    appDescriptionLabel.setText(
        "".join(
            filter(str.isprintable, appDescriptionLabel.text())
        )
    )

def get_duration(duration_idx):
    if duration_idx == 0:
        return Config.DURATION_ONCE
    elif duration_idx == 1:
        return _constants.DURATION_30s
    elif duration_idx == 2:
        return _constants.DURATION_5m
    elif duration_idx == 3:
        return _constants.DURATION_15m
    elif duration_idx == 4:
        return _constants.DURATION_30m
    elif duration_idx == 5:
        return _constants.DURATION_1h
    elif duration_idx == 6:
        return _constants.DURATION_12h
    elif duration_idx == 7:
        return Config.DURATION_UNTIL_RESTART
    else:
        return Config.DURATION_ALWAYS

def set_default_duration(cfg, durationCombo):
    if cfg.hasKey(Config.DEFAULT_DURATION_KEY):
        cur_idx = cfg.getInt(Config.DEFAULT_DURATION_KEY)
        durationCombo.setCurrentIndex(cur_idx)
    else:
        durationCombo.setCurrentIndex(Config.DEFAULT_DURATION_IDX)

def set_default_target(combo, con, cfg, app_name, app_args):
    # set appimage as default target if the process path starts with
    # /tmp/._mount
    if con.process_path.startswith(_constants.APPIMAGE_PREFIX):
        idx = combo.findData(_constants.FIELD_APPIMAGE)
        if idx != -1:
            combo.setCurrentIndex(idx)
            return

    if int(con.process_id) > 0 and app_name != "" and app_args != "":
        combo.setCurrentIndex(int(cfg.getSettings(cfg.DEFAULT_TARGET_KEY)))
    else:
        combo.setCurrentIndex(2)

def get_combo_operator(data, comboText, con):
    if data == _constants.FIELD_PROC_PATH:
        return Config.RULE_TYPE_SIMPLE, Config.OPERAND_PROCESS_PATH, con.process_path

    elif data == _constants.FIELD_PROC_ARGS:
        # this should not happen
        if len(con.process_args) == 0 or con.process_args[0] == "":
            return Config.RULE_TYPE_SIMPLE, Config.OPERAND_PROCESS_PATH, con.process_path
        return Config.RULE_TYPE_SIMPLE, Config.OPERAND_PROCESS_COMMAND, ' '.join(con.process_args)

    elif data == _constants.FIELD_PROC_ID:
        return Config.RULE_TYPE_SIMPLE, Config.OPERAND_PROCESS_ID, "{0}".format(con.process_id)

    elif data == _constants.FIELD_USER_ID:
        return Config.RULE_TYPE_SIMPLE, Config.OPERAND_USER_ID, "{0}".format(con.user_id)

    elif data == _constants.FIELD_DST_PORT:
        return Config.RULE_TYPE_SIMPLE, Config.OPERAND_DEST_PORT, "{0}".format(con.dst_port)

    elif data == _constants.FIELD_DST_IP:
        return Config.RULE_TYPE_SIMPLE, Config.OPERAND_DEST_IP, con.dst_ip

    elif data == _constants.FIELD_DST_HOST:
        return Config.RULE_TYPE_SIMPLE, Config.OPERAND_DEST_HOST, comboText

    elif data == _constants.FIELD_DST_NETWORK:
        # strip "to ": "to x.x.x/20" -> "x.x.x/20"
        # we assume that to is one word in all languages
        parts = comboText.split(' ')
        text = parts[len(parts)-1]
        return Config.RULE_TYPE_NETWORK, Config.OPERAND_DEST_NETWORK, text

    elif data == _constants.FIELD_REGEX_HOST:
        parts = comboText.split(' ')
        text = parts[len(parts)-1]
        # ^(|.*\.)yahoo\.com
        dsthost = r'\.'.join(text.split('.')).replace("*", "")
        dsthost = r'^(|.*\.)%s$' % dsthost[2:]
        return Config.RULE_TYPE_REGEXP, Config.OPERAND_DEST_HOST, dsthost

    elif data == _constants.FIELD_REGEX_IP:
        parts = comboText.split(' ')
        text = parts[len(parts)-1]
        return Config.RULE_TYPE_REGEXP, Config.OPERAND_DEST_IP, "%s" % r'\.'.join(text.split('.')).replace("*", ".*")

    elif data == _constants.FIELD_APPIMAGE:
        appimage_bin = os.path.basename(con.process_path)
        appimage_path = os.path.dirname(con.process_path).replace('.', r'\.')
        appimage_path = appimage_path[0:len(_constants.APPIMAGE_PREFIX)+7]
        return Config.RULE_TYPE_REGEXP, Config.OPERAND_PROCESS_PATH, r'^{0}[0-9A-Za-z]{{6,7}}\/.*{1}$'.format(appimage_path, appimage_bin)



from PyQt5.QtCore import QCoreApplication as QC
from opensnitch.config import Config
from opensnitch.rules import Rule

def verify(checksums, rule):
    """return true if the checksum of a rule matches the one of the process
    opening a connection.
    """
    # when verifying checksums, we'll always have a rule type List, with at
    # least: path of the process + checksum
    if rule.operator.type != Config.RULE_TYPE_LIST:
        return True, ""

    # checksum will be empty if the daemon failed to calculate it.
    # in this case assume that it's ok (ignore it).
    if checksums[Config.OPERAND_PROCESS_HASH_MD5] == "":
        return True, ""

    if not rule.enabled:
        return True, ""
    for ro in rule.operator.list:
        if ro.type == Config.RULE_TYPE_SIMPLE and ro.operand == Config.OPERAND_PROCESS_HASH_MD5:
            if ro.data != checksums[Config.OPERAND_PROCESS_HASH_MD5]:
                return False, ro.data

    return True, ""

def update_rule(node, rules, rule_name, con):
    """try to obtain the rule from the DB by name.
    return the rule on success, or None + error message on error.
    """

    # get rule from the db
    records = rules.get_by_name(node, rule_name)
    if records == None or records.first() == False:
        return None, QC.translate("popups", "Rule not updated, not found by name ({0})".format(rule_name))

    # transform it to proto rule
    rule_obj = Rule.new_from_records(records)
    if rule_obj.operator.type != Config.RULE_TYPE_LIST:
        if rule_obj.operator.operand == Config.OPERAND_PROCESS_HASH_MD5:
            rule_obj.operator.data = con.process_checksums[Config.OPERAND_PROCESS_HASH_MD5]
    else:
        for op in rule_obj.operator.list:
            if op.operand == Config.OPERAND_PROCESS_HASH_MD5:
                op.data = con.process_checksums[Config.OPERAND_PROCESS_HASH_MD5]
                break
    # add it back again to the db
    added = rules.add_rules(node, [rule_obj])
    if not added:
        return None, QC.translate("popups", "Rule not updated ({0}).".format(rule_name))

    return rule_obj, ""



from PyQt5.QtCore import QCoreApplication as QC

PAGE_MAIN = 2
PAGE_DETAILS = 0
PAGE_CHECKSUMS = 1

DEFAULT_TIMEOUT = 15

# don't translate
FIELD_REGEX_HOST    = "regex_host"
FIELD_REGEX_IP      = "regex_ip"
FIELD_PROC_PATH     = "process_path"
FIELD_PROC_ARGS     = "process_args"
FIELD_PROC_ID       = "process_id"
FIELD_USER_ID       = "user_id"
FIELD_DST_IP        = "dst_ip"
FIELD_DST_PORT      = "dst_port"
FIELD_DST_NETWORK   = "dst_network"
FIELD_DST_HOST      = "simple_host"
FIELD_APPIMAGE      = "appimage_path"

DURATION_30s    = "30s"
DURATION_5m     = "5m"
DURATION_15m    = "15m"
DURATION_30m    = "30m"
DURATION_1h     = "1h"
DURATION_12h     = "12h"
# don't translate

APPIMAGE_PREFIX = "/tmp/.mount_"

# label displayed in the pop-up combo
DURATION_session = QC.translate("popups", "until reboot")
# label displayed in the pop-up combo
DURATION_forever = QC.translate("popups", "forever")



import json

# https://doc.qt.io/qtforpython-5/overviews/richtext-html-subset.html#supported-html-subset
# XXX: this is just a quick way of beautifying the json response. PRs and
# suggestions to improve it are welcome! like displaying a summary of the
# results, a link to the VT web, etc.
def build_html_report(json_obj, report="", spaces=""):
    for i in json_obj:
        if type(json_obj[i]) == dict:
            report = report + spaces + "<ul><strong>" + i + "</strong>\n"
            report = build_html_report(json_obj[i], report, spaces+"")
            report = report + "</ul>\n"
        elif type(json_obj[i]) == str:
            report = report + spaces + "<li><strong>" + i + ":</strong>  " + json_obj[i].replace("\n", "<br>") + "</li>\n"
        elif type(json_obj[i]) == int or type(json_obj[i]) == bool:
            report = report + spaces + "<li><strong>{0}:</strong>  {1}</li>\n".format(i, json_obj[i])
        elif type(json_obj[i]) == list:
            report = report + spaces + "\n<p><strong>" + i + "</strong>: \n"
            li = ""
            for l in json_obj[i]:
                li = "{0}{1},".format(li, l)
            report = report + spaces + li + "</p>\n"
        elif type(i) == str:
            report = report + spaces + "<ul><strong>" + i + "</strong>: "
            if type(json_obj[i]) == list:
                for l in json_obj[i]:
                    report = "<li>{0}{1}</li>\n".format(report, l)
            else:
                report = "<li>{0}{1}</li>\n".format(report, json_obj[i])
            report = report + spaces + "</ul>\n"

    return report


def report_to_html(json_obj, report="", spaces=""):
    # https://www.virustotal.com/gui/domain/malware.wicar.org
    rpt = build_html_report(json_obj, report, spaces)
    # FIXME: lists indentation
    return "<html><style>ul {list-style: none;}</style><body>" + rpt + "</body></html>"

def report_to_ascii(json_obj, report="", space=""):
    for i in json_obj:
        if type(json_obj[i]) == dict:
            report = report + "\n" + space + i + "\n"
            report = report_to_ascii(json_obj[i], report, space+"    ")
        elif type(json_obj[i]) == str:
            report = report + space + i + ":  " + json_obj[i] + "\n"
        elif type(i) == str:
            report = report + space + " - " + i + ":  "
            if type(json_obj[i]) == list:
                for l in json_obj[i]:
                    report = "{0}{1}\n".format(report, l)
            else:
                report = "{0}{1}\n".format(report, json_obj[i])

    return report



from opensnitch.utils import Enums
from opensnitch.config import Config

class Verdicts(Enums):
    EMPTY = ""
    ACCEPT = Config.ACTION_ACCEPT
    DROP = Config.ACTION_DROP
    REJECT = Config.ACTION_REJECT
    RETURN = Config.ACTION_RETURN
    QUEUE = Config.ACTION_QUEUE
    DNAT = Config.ACTION_DNAT
    SNAT = Config.ACTION_SNAT
    REDIRECT = Config.ACTION_REDIRECT
    TPROXY = Config.ACTION_TPROXY
    #MASQUERADE = Config.ACTION_MASQUERADE
    #LOG = Config.ACTION_LOG
    STOP = Config.ACTION_STOP



class Policy(Enums):
    ACCEPT = "accept"
    DROP = "drop"

class Table(Enums):
    FILTER = "filter"
    MANGLE = "mangle"
    NAT = "nat"

class Hooks(Enums):
    INPUT  ="input"
    OUTPUT  ="output"
    FORWARD = "forward"
    PREROUTING = "prerouting"
    POSTROUTING = "postrouting"

class PortProtocols(Enums):
    TCPUDP = "tcp,udp"
    TCP = "tcp"
    UDP = "udp"
    UDPLITE = "udplite"
    SCTP = "sctp"
    DCCP = "dccp"

class Protocols(Enums):
    TCP = "tcp"
    UDP = "udp"
    UDPLITE = "udplite"
    SCTP = "sctp"
    DCCP = "dccp"
    ICMP = "icmp"
    ICMPv6 = "icmpv6"
    AH = "ah"
    ETHERNET = "ethernet"
    GREP = "gre"
    IP = "ip"
    IPIP = "ipip"
    L2TP = "l2tp"
    COMP = "comp"
    IGMP = "igmp"
    ESP = "esp"
    RAW = "raw"
    ENCAP = "encap"

class Family(Enums):
    INET = "inet"
    IPv4 = "ip"
    IPv6 = "ip6"

class ChainType(Enums):
    FILTER = "filter"
    MANGLE = "mangle"
    ROUTE = "route"
    SNAT = "natsource"
    DNAT = "natdest"

class Operator(Enums):
    EQUAL = "=="
    NOT_EQUAL = "!="
    GT_THAN = ">="
    GT = ">"
    LT_THAN = "<="
    LT = "<"

class TimeUnits(Enums):
    SECOND = "second"
    MINUTE = "minute"
    HOUR = "hour"
    DAY = "day"

class RateUnits(Enums):
    BYTES = "bytes"
    KBYTES = "kbytes"
    MBYTES = "mbytes"
    GBYTES = "gbytes"

class Statements(Enums):
    """Enum of known (allowed) statements:
        [tcp,udp,ip] ...
    """
    # we may need in the future:
    # ANY = tcp,udp,udplite,sctp,dccp
    TCPUDP = "tcp,udp"
    TCP = "tcp"
    UDP = "udp"
    UDPLITE = "udplite"
    SCTP = "sctp"
    DCCP = "dccp"
    ICMP = "icmp"
    ICMPv6 = "icmpv6"

    SPORT = "sport"
    DPORT = "dport"
    DADDR = "daddr"
    SADDR = "saddr"

    IP = "ip"
    IP6 = "ip6"
    IIFNAME = "iifname"
    OIFNAME = "oifname"
    CT = "ct"
    META = "meta"
    COUNTER = "counter"
    NAME = "name"
    LOG = "log"
    QUOTA = "quota"
    LIMIT = "limit"



from PyQt5 import QtGui
from opensnitch.config import Config

def render(node, detailsWidget, con):
    tree = ""
    space = "&nbsp;"
    spaces = "&nbsp;"
    indicator = ""

    try:
        # reverse() doesn't exist on old protobuf libs.
        con.process_tree.reverse()
    except:
        pass
    for path in con.process_tree:
        tree = "{0}<p>│{1}\t{2}{3}{4}</p>".format(tree, path.value, spaces, indicator, path.key)
        spaces += "&nbsp;" * 4
        indicator = "\\_ "

    # XXX: table element doesn't work?
    details = """<b>{0}</b> {1}:{2} -> {3}:{4}
<br><br>
<b>Path:</b>{5}{6}<br>
<b>Cmdline:</b>&nbsp;{7}<br>
<b>CWD:</b>{8}{9}<br>
<b>MD5:</b>{10}{11}<br>
<b>UID:</b>{12}{13}<br>
<b>PID:</b>{14}{15}<br>
<br>
<b>Process tree:</b><br>
{16}
<br>
<p><b>Environment variables:<b></p>
{17}
""".format(
con.protocol.upper(),
con.src_port, con.src_ip, con.dst_ip, con.dst_port,
space * 6, con.process_path,
" ".join(con.process_args),
space * 6, con.process_cwd,
space * 7, con.process_checksums[Config.OPERAND_PROCESS_HASH_MD5],
space * 9, con.user_id,
space * 9, con.process_id,
tree,
"".join('<p>{}={}</p>'.format(key, value) for key, value in con.process_env.items())
)

    detailsWidget.document().clear()
    detailsWidget.document().setHtml(details)
    detailsWidget.moveCursor(QtGui.QTextCursor.Start)



from PyQt5.QtCore import QObject, pyqtSignal

from opensnitch.database import Database
from opensnitch.database.enums import RuleFields
from opensnitch.config import Config

import opensnitch.proto as proto
ui_pb2, ui_pb2_grpc = proto.import_()

import os
import json
from slugify import slugify
from datetime import datetime
from google.protobuf.json_format import MessageToJson, Parse

DefaultRulesPath = "/etc/opensnitchd/rules"

# date format displayed on the GUI (created column)
DBDateFieldFormat = "%Y-%m-%d %H:%M:%S"

class Rule():
    def __init__(self):
        pass

    @staticmethod
    def to_bool(s):
        return s == 'True'

    @staticmethod
    def new_empty():
        pass

    @staticmethod
    def new_from_records(records):
        """Creates a new protobuf Rule from DB records.
        Fields of the record are in the order defined on the DB.
        """
        rule = ui_pb2.Rule(name=records.value(RuleFields.Name))
        rule.enabled = Rule.to_bool(records.value(RuleFields.Enabled))
        rule.precedence = Rule.to_bool(records.value(RuleFields.Precedence))
        rule.action = records.value(RuleFields.Action)
        rule.duration = records.value(RuleFields.Duration)
        rule.operator.type = records.value(RuleFields.OpType)
        rule.operator.sensitive = Rule.to_bool(records.value(RuleFields.OpSensitive))
        rule.operator.operand = records.value(RuleFields.OpOperand)
        rule.operator.data = "" if records.value(RuleFields.OpData) == None else str(records.value(RuleFields.OpData))
        rule.description = records.value(RuleFields.Description)
        rule.nolog = Rule.to_bool(records.value(RuleFields.NoLog))
        created = int(datetime.now().timestamp())
        if records.value(RuleFields.Created) != "":
            created = int(datetime.strptime(
                records.value(RuleFields.Created), DBDateFieldFormat
            ).timestamp())
        rule.created = created

        try:
            # Operator list is always saved as json string to the db,
            # so we need to load the json string.
            if rule.operator.type == Config.RULE_TYPE_LIST:
                operators = json.loads(rule.operator.data)
                for op in operators:
                    rule.operator.list.extend([
                        ui_pb2.Operator(
                            type=op['type'],
                            operand=op['operand'],
                            sensitive=False if op.get('sensitive') == None else op['sensitive'],
                            data="" if op.get('data') == None else op['data']
                        )
                    ])
                rule.operator.data = ""
        except Exception as e:
            print("new_from_records exception parsing operartor list:", e)


        return rule

class Rules(QObject):
    __instance = None
    updated = pyqtSignal(int)

    LOG_TAG = "[Rules]: "

    @staticmethod
    def instance():
        if Rules.__instance == None:
            Rules.__instance = Rules()
        return Rules.__instance

    def __init__(self):
        QObject.__init__(self)
        self._db = Database.instance()

    def add(self, time, node, name, description, enabled, precedence, nolog, action, duration, op_type, op_sensitive, op_operand, op_data, created):
        # don't add rule if the user has selected to exclude temporary
        # rules
        if duration in Config.RULES_DURATION_FILTER:
            return

        self._db.insert("rules",
                  "(time, node, name, description, enabled, precedence, nolog, action, duration, operator_type, operator_sensitive, operator_operand, operator_data, created)",
                  (time, node, name, description, enabled, precedence, nolog, action, duration, op_type, op_sensitive, op_operand, op_data, created),
                        action_on_conflict="REPLACE")

    def add_rules(self, addr, rules):
        try:
            for _,r in enumerate(rules):
                # Operator list is always saved as json string to the db.
                rjson = json.loads(MessageToJson(r))
                if r.operator.type == Config.RULE_TYPE_LIST and \
                        rjson.get('operator') != None and \
                        rjson.get('operator').get('list') != None:

                    r.operator.data = json.dumps(rjson.get('operator').get('list'))

                self.add(datetime.now().strftime(DBDateFieldFormat),
                         addr,
                         r.name, r.description, str(r.enabled),
                         str(r.precedence), str(r.nolog), r.action, r.duration,
                         r.operator.type,
                         str(r.operator.sensitive),
                         r.operator.operand, r.operator.data,
                         str(datetime.fromtimestamp(r.created).strftime(DBDateFieldFormat)))

            return True
        except Exception as e:
            print(self.LOG_TAG + " exception adding node rules to db: ", e)
            return False

    def delete(self, name, addr, callback):
        rule = ui_pb2.Rule(name=name)
        rule.enabled = False
        rule.action = ""
        rule.duration = ""
        rule.operator.type = ""
        rule.operator.operand = ""
        rule.operator.data = ""

        if not self._db.delete_rule(rule.name, addr):
            return None

        return rule

    def delete_by_field(self, field, values):
        return self._db.delete_rules_by_field(field, values)

    def get_by_name(self, node, name):
        return self._db.get_rule(name, node)

    def get_by_field(self, node, field, value):
        return self._db.get_rule_by_field(node, field, value)

    def exists(self, rule, node_addr):
        return self._db.rule_exists(rule, node_addr)

    def new_unique_name(self, rule_name, node_addr, prefix):
        """generate a new name, if the supplied one already exists
        """
        if self._db.get_rule(rule_name, node_addr).next() == False:
            return rule_name

        for idx in range(0, 100):
            new_rule_name = "{0}-{1}".format(rule_name, idx)
            if self._db.get_rule(new_rule_name, node_addr).next() == False:
                return new_rule_name

        return rule_name

    def disable(self, addr, name):
        """Mark a rule as not enabled in the DB"""
        self._db.update(
            "rules",
            "enabled='False'",
            (name, addr),
            "name=? AND node=?",
            action_on_conflict="OR REPLACE"
        )

    def update_time(self, time, name, addr):
        """Updates the time of a rule, whenever a new connection matched a
        rule.
        """
        self._db.update("rules",
                        "time=?",
                        (time, name, addr),
                        "name=? AND node=?",
                        action_on_conflict="OR REPLACE"
                        )

    def _timestamp_to_rfc3339(self, time):
        """converts timestamp to rfc3339 format"""
        return "{0}Z".format(
            datetime.fromtimestamp(time).isoformat(timespec='microseconds')
        )

    def rule_to_json(self, node, rule_name):
        try:
            records = self._db.get_rule(rule_name, node)
            if records == None or records == -1:
                return None
            if not records.next():
                return None
            rule = Rule.new_from_records(records)
            # exclude this field when exporting to json
            tempRule = MessageToJson(rule)
            jRule = json.loads(tempRule)
            jRule['created'] = self._timestamp_to_rfc3339(rule.created)
            return json.dumps(jRule, indent="    ")
        except Exception as e:
            print("rule_to_json() exception:", e)
            return None

    def _export_rule_common(self, node, records, outdir):
        try:
            rule = Rule.new_from_records(records)
            rulename = rule.name
            if ".json" not in rulename:
                rulename = rulename + ".json"
            with open(outdir  + "/" + rulename, 'w') as jsfile:
                actual_json_text = MessageToJson(rule)
                jRule = json.loads(actual_json_text)
                jRule['created'] = self._timestamp_to_rfc3339(rule.created)
                actual_json_text = json.dumps(jRule, indent="    ")
                jsfile.write( actual_json_text )

            return True
        except Exception as e:
            print(self.LOG_TAG, "export_rules(", node, outdir, ") exception:", e)

        return False

    def export_rule(self, node, rule_name, outdir):
        """Gets the rule from the DB and writes it out to a directory.
        A new directory per node will be created.
        """
        try:
            records = self._db.get_rule(rule_name, node)
            if records.next() == False:
                print("export_rule() get_error 2:", records)
                return False

            rulesdir = outdir + "/" + slugify(node)
            try:
                os.makedirs(rulesdir, 0o700)
            except Exception as e:
                print("exception creating dirs:", e)

            return self._export_rule_common(node, records, rulesdir)

        except Exception as e:
            print(self.LOG_TAG, "export_rules(", node, rulesdir, ") exception:", e)

        return False

    def export_rules(self, node, outdir):
        """Gets the rules from the DB and writes them out to a directory.
        A new directory per node will be created.
        """
        records = self._db.get_rules(node)
        if records == None:
            return False

        rulesdir = outdir + "/" + slugify(node)
        try:
            os.makedirs(rulesdir, 0o700)
        except Exception as e:
            print("exception creating dirs:", e)
        try:
            while records.next() != False:
                self._export_rule_common(node, records, rulesdir)

        except Exception as e:
            print(self.LOG_TAG, "export_rules(", node, rulesdir, ") exception:", e)
            return False

        return True

    def import_rules(self, rulesdir):
        """Read a directory with rules in json format, and parse them out to protobuf
        Returns a list of rules on success, or None on error.
        """
        try:
            rules = []
            for rulename in os.listdir(rulesdir):
                with open(rulesdir  + "/" + rulename, 'r') as f:
                    jsrule = f.read()
                    # up until v1.6.5/v1.7.0, 'created' field was exported as timestamp.
                    # since > v1.6.5 it's exported in rfc3339 format, so if we fail to
                    # parse the rule, we'll try to convert the 'created' value from
                    # timestamp to rfc3339.
                    try:
                        pb_rule = Parse(text=jsrule, message=ui_pb2.Rule(), ignore_unknown_fields=True)
                    except:
                        jRule = json.loads(jsrule)
                        created = int(datetime.strptime(
                            jRule['created'], "%Y-%m-%dT%H:%M:%S.%fZ"
                        ).timestamp())
                        jRule['created'] = created
                        jsrule = json.dumps(jRule)
                        pb_rule = Parse(text=jsrule, message=ui_pb2.Rule(), ignore_unknown_fields=True)
                    rules.append(pb_rule)

            return rules
        except Exception as e:
            print(self.LOG_TAG, "import_rules() exception:", e)

        return None



import os
import os.path
import logging
import requests
import threading
from queue import Queue

from opensnitch.version import version
from opensnitch.dialogs.stats import StatsDialog
from opensnitch.notifications import DesktopNotifications
from opensnitch.plugins import PluginBase, PluginSignal
from opensnitch.utils.xdg import xdg_config_home
from opensnitch.utils import GenericTimer

ch = logging.StreamHandler()
#ch.setLevel(logging.ERROR)
formatter = logging.Formatter('%(asctime)s - %(name)s - [%(levelname)s] %(message)s')
ch.setFormatter(formatter)
logger = logging.getLogger(__name__)
logger.addHandler(ch)
logger.setLevel(logging.WARNING)

class Downloader(PluginBase):
    """A plugin that schedules downloads from remote urls.

    This plugin may require to create a rule to allow connections to the
    configured urls, to avoid popups.
    """
    # fields overriden from parent class
    name = "Downloader"
    version = 0
    author = "opensnitch"
    created = ""
    modified = ""
    enabled = False
    description = "Download remote resources to local directories"

    # default
    TYPE = [PluginBase.TYPE_GLOBAL]

    # list of scheduled tasks
    scheduled_tasks = {}
    default_conf = "{0}/{1}".format(xdg_config_home, "/opensnitch/actions/downloaders.json")

    def __init__(self, config=None):
        self.signal_in.connect(self.cb_signal)
        self._config = config
        self._desktop_notifications = DesktopNotifications()
        self._ok_msg = ""
        self._err_msg = ""
        self._notify_title = "[OpenSnitch] Blocklist downloader"
        self._resultsQueue = Queue()
        self._app_icon = os.path.join(os.path.abspath(os.path.dirname(__file__)), "../../res/icon-white.svg")

    def configure(self, parent=None):
        # TODO:
        if type(parent) == StatsDialog:
            pass
            #_gui.add_panel_section()

    def compile(self):
        """Transform a json object to python objects.
        """
        logger.debug("compile()")
        try:
            if self._config.get('config') == None:
                logger.warning("compile() config:[] missing")
                return

            settings = self._config['config']
            for idx, config in enumerate(settings):
                for url in config['urls']:
                    if url['localfile'] == "" or url['remote'] == "" or config['name'] == "":
                        logger.debug("compile() downloader name, url, localfile, units and interval fields cannot be empty")
                        continue
                    localdir = os.path.dirname(url['localfile'])
                    if not os.path.isdir(localdir):
                        os.makedirs(localdir, 0o700)

                self._notify = config.get('notify')
                if self._notify != None:
                    ok = self._notify.get('success')
                    err = self._notify.get('error')
                    if ok != None:
                        ok_msg = ok.get('desktop')
                        if ok_msg:
                            self._ok_msg = ok_msg
                    if err != None:
                        err_msg = err.get('desktop')
                        if err_msg:
                            self._err_msg = err_msg

                if config['units'] == 'seconds':
                    interval = float(config['interval'])
                elif config['units'] == 'minutes':
                    interval = float(config['interval']) * 60
                elif config['units'] == 'hours':
                    interval = (float(config['interval']) * 60) * 60
                elif config['units'] == 'days':
                    interval = ((float(config['interval']) * 60) * 60) * 60
                else:
                    logger.warning("compile() unknown time format '{0}'".format(config['units']))

                self.scheduled_tasks[config['name']] = self.new_timer(interval, config)
        except Exception as e:
            logger.warning("compile() exception:", extra=e)

    def run(self, parent=None, args=()):
        """Run the action on the given arguments.
        """
        if parent == StatsDialog:
            pass

        settings = self._config['config']
        for idx, config in enumerate(settings):
            self.scheduled_tasks[config['name']].start()

    def new_timer(self, interval, config):
        logger.debug("new_timer()")
        t = GenericTimer(interval, True, self.cb_run_tasks, (self.scheduled_tasks, config,))
        return t

    def cb_run_tasks(self, args):
        try:
            tasks, config = args
            updated = True
            failed_urls = []
            threads = []
            for url in config['urls']:
                if not url['enabled']:
                    logger.debug("cb_run_tasks() disabled -> %s", url['remote'])
                    continue

                # TODO:
                #  - check and save content-length header, to avoid unnecessary
                #  downloads.
                #  - cancel on too many failures
                #  - on slow connections/too much urls + refresh interval too
                #  low, timers may overlap.
                #
                #response = requests.head(url['remote'])
                #print("Downloaders >> {0} >> length >> {1}".format(url, response.headers['content-length']))
                th = threading.Thread(
                        target=self.download_url,
                        args=(config['name'], url['remote'], url['localfile'],)
                    )
                th.start()
                threads.append(th)

            for th in threads:
                th.join()

            while not self._resultsQueue.empty():
                remote, ok = self._resultsQueue.get_nowait()
                updated &= ok
                if not ok:
                    failed_urls.append(remote)

            result_msg = self._ok_msg if updated else self._err_msg
            result_msg = "{0}\n\n{1}".format(result_msg, failed_urls if len(failed_urls) > 0 else "")
            if self._notify != None:
                if self._desktop_notifications.is_available() and self._desktop_notifications.are_enabled():
                    self._desktop_notifications.show(
                        self._notify_title,
                        result_msg,
                        self._app_icon
                    )
                else:
                    logger.debug("notification module is not available or disabled.")

        except Exception as e:
            logger.warning("cb_run_tasks.exception: %s", repr(e))

    def cb_signal(self, signal):
        logger.debug("cb_signal: %s, %s", self.name, signal)
        try:
            if signal == PluginSignal.ENABLE:
                self.enabled = True

            if signal['signal'] == PluginSignal.DISABLE or signal['signal'] == PluginSignal.STOP:
                for t in self.scheduled_tasks:
                    logger.debug("cb_signal.stopping task: %s, %s", self.name, signal)
                    self.scheduled_tasks[t].stop()

        except Exception as e:
            logger.warning("cb_signal() exception: %s", repr(e))

    def download_url(self, task_name, remote, localfile):
        logger.debug("download_url() %s -> %s", remote, localfile)

        try:
            response = requests.get(
                remote,
                # introduced in v0.12.0 (2012)
                # XXX: support prefetch= ?
                stream=True,
                headers={'User-Agent': 'OpenSnitch v%s' % version}
            )
            if response.status_code != 200:
                logger.debug("download_url() error: %s", task_name)
                self._resultsQueue.put((remote, False))
                return False

            with open(localfile, "wb") as f:
                # write content in chunks, to avoid excessive mem usage.
                for chunk in response.iter_content(1024):
                    if not chunk:
                        break
                    f.write(chunk)

        except Exception as e:
            logger.debug("download_url() exception: %s", repr(e))
            self._resultsQueue.put((remote, False))
            return False

        self._resultsQueue.put((remote, True))
        return True



import json
from PyQt5 import QtWidgets
from opensnitch.plugins.virustotal.virustotal import VTAnalysis, Virustotal
from opensnitch.plugins.virustotal import _utils

def build_vt_tab(plugin, parent):
    """add a new tab with a text field that will contain the result of the query in JSON format.
    """
    wdg_count = parent.tabWidget.count()
    prev_wdg = parent.tabWidget.widget(wdg_count)
    if prev_wdg != None and prev_wdg.objectName() == "vt_tab":
        return prev_wdg

    wdg = QtWidgets.QWidget()
    wdg.setObjectName("vt_tab")
    gridLayout =  QtWidgets.QGridLayout()
    textWdg = QtWidgets.QTextEdit()
    textWdg.setSizePolicy(QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Expanding)
    gridLayout.addWidget(textWdg, 0, 0)
    wdg.setLayout(gridLayout)

    return wdg

def add_vt_tab(vt, parent, widget):
    parent.tabWidget.addTab(widget, "Virustotal")
    parent.tabWidget.currentChanged.connect(lambda idx: _cb_proctab_changed(idx, vt, parent))

def _cb_proctab_changed(idx, vt, parent):
    cur_tab = parent.tabWidget.widget(idx)
    if cur_tab.objectName() != "vt_tab":
        return

    tmp = parent.labelChecksums.text().split(" ")
    lblChecksum = ""
    if len(tmp) > 1:
        lblChecksum = tmp[1]

    if lblChecksum == "":
        return

    url = vt.API_FILES + lblChecksum
    vt_thread = VTAnalysis(parent, vt._config, "", url, vt.API_CONNECT_TIMEOUT, vt.API_KEY, None)
    vt_thread.signals.completed.connect(vt.analysis_completed)
    vt_thread.signals.error.connect(vt.analysis_error)
    vt.threadsPool.start(vt_thread)

def update_tab(what, response, parent, config, conn):
    tabs = parent.tabWidget.count()-1
    cur_tab = parent.tabWidget.widget(tabs)
    if cur_tab.objectName() != "vt_tab":
        return
    result = json.loads(response.content)
    textWdg = cur_tab.layout().itemAtPosition(0, 0).widget()
    textWdg.clear()

    if result.get('data') == None:
        textWdg.setPlainText("checksum not found on Virustotal. Upload the binary for analysis and generate a report.")
        return
    textWdg.setHtml(_utils.report_to_html(result))



import threading
import sys
import time
import os
import os.path
import pwd
import json
import ipaddress
from datetime import datetime

from PyQt5 import QtCore, uic, QtWidgets
from PyQt5.QtCore import QCoreApplication as QC, QEvent

from slugify import slugify

from opensnitch.utils import Icons
from opensnitch.desktop_parser import LinuxDesktopParser
from opensnitch.config import Config
from opensnitch.version import version
from opensnitch.actions import Actions
from opensnitch.plugins import PluginBase
from opensnitch.rules import Rules, Rule
from opensnitch.nodes import Nodes

from opensnitch.dialogs.prompt import _utils, _constants, _checksums, _details
import opensnitch.proto as proto
ui_pb2, ui_pb2_grpc = proto.import_()

from opensnitch.utils.network_aliases import NetworkAliases

DIALOG_UI_PATH = "%s/../../res/prompt.ui" % os.path.dirname(sys.modules[__name__].__file__)
class PromptDialog(QtWidgets.QDialog, uic.loadUiType(DIALOG_UI_PATH)[0]):
    _prompt_trigger = QtCore.pyqtSignal()
    _tick_trigger = QtCore.pyqtSignal()
    _timeout_trigger = QtCore.pyqtSignal()

    TYPE = "popups"

    def __init__(self, parent=None, appicon=None):
        QtWidgets.QDialog.__init__(self, parent, QtCore.Qt.WindowStaysOnTopHint)
        # Other interesting flags: QtCore.Qt.Tool | QtCore.Qt.BypassWindowManagerHint
        self._cfg = Config.get()
        self._rules = Rules.instance()
        self._nodes = Nodes.instance()

        self.setupUi(self)
        self.setWindowIcon(appicon)
        self.installEventFilter(self)

        self._width = None
        self._height = None

        dialog_geometry = self._cfg.getSettings("promptDialog/geometry")
        if dialog_geometry == QtCore.QByteArray:
            self.restoreGeometry(dialog_geometry)

        self.setWindowTitle("OpenSnitch v%s" % version)

        self._actions = Actions.instance()
        self._action_list = self._actions.getByType(PluginBase.TYPE_POPUPS)
        self._configure_plugins()

        self._lock = threading.Lock()
        self._con = None
        self._rule = None
        self._local = True
        self._peer = None
        self._prompt_trigger.connect(self.on_connection_prompt_triggered)
        self._timeout_trigger.connect(self.on_timeout_triggered)
        self._tick_trigger.connect(self.on_tick_triggered)
        self._tick = int(self._cfg.getSettings(self._cfg.DEFAULT_TIMEOUT_KEY)) if self._cfg.hasKey(self._cfg.DEFAULT_TIMEOUT_KEY) else _constants.DEFAULT_TIMEOUT
        self._tick_thread = None
        self._done = threading.Event()
        self._timeout_text = ""
        self._timeout_triggered = False

        self._apps_parser = LinuxDesktopParser()

        self.whatIPCombo.setVisible(False)
        self.checkDstIP.setVisible(False)
        self.checkDstPort.setVisible(False)
        self.checkUserID.setVisible(False)
        self.appDescriptionLabel.setVisible(False)

        self._ischeckAdvanceded = False
        self.checkAdvanced.toggled.connect(self._check_advanced_toggled)

        self.checkAdvanced.clicked.connect(self._button_clicked)
        self.durationCombo.activated.connect(self._button_clicked)
        self.whatCombo.activated.connect(self._button_clicked)
        self.whatIPCombo.activated.connect(self._button_clicked)
        self.checkDstIP.clicked.connect(self._button_clicked)
        self.checkDstPort.clicked.connect(self._button_clicked)
        self.checkUserID.clicked.connect(self._button_clicked)
        self.cmdInfo.clicked.connect(self._cb_cmdinfo_clicked)
        self.cmdBack.clicked.connect(self._cb_cmdback_clicked)

        self.cmdUpdateRule.clicked.connect(lambda: self._cb_update_rule_clicked(updateAll=False))
        self.cmdUpdateRuleAll.clicked.connect(lambda: self._cb_update_rule_clicked(updateAll=True))
        self.cmdBackChecksums.clicked.connect(self._cb_cmdback_clicked)
        self.messageLabel.linkActivated.connect(self._cb_warninglbl_clicked)

        self.allowIcon = Icons.new(self, "emblem-default")
        denyIcon = Icons.new(self, "emblem-important")
        rejectIcon = Icons.new(self, "window-close")
        backIcon = Icons.new(self, "go-previous")
        infoIcon = Icons.new(self, "dialog-information")

        self.cmdInfo.setIcon(infoIcon)
        self.cmdBack.setIcon(backIcon)
        self.cmdBackChecksums.setIcon(backIcon)

        self._default_action = self._cfg.getInt(self._cfg.DEFAULT_ACTION_KEY)

        self.allowButton.clicked.connect(lambda: self._on_action_clicked(Config.ACTION_ALLOW_IDX))
        self.allowButton.setIcon(self.allowIcon)
        self._allow_text = QC.translate("popups", "Allow")
        self._action_text = [
            QC.translate("popups", "Deny"),
            QC.translate("popups", "Allow"),
            QC.translate("popups", "Reject")
        ]
        self._action_icon = [denyIcon, self.allowIcon, rejectIcon]

        m = QtWidgets.QMenu()
        m.addAction(denyIcon, self._action_text[Config.ACTION_DENY_IDX]).triggered.connect(
            lambda: self._on_action_clicked(Config.ACTION_DENY_IDX)
        )
        m.addAction(rejectIcon, self._action_text[Config.ACTION_REJECT_IDX]).triggered.connect(
            lambda: self._on_action_clicked(Config.ACTION_REJECT_IDX)
        )
        self.actionButton.setMenu(m)
        self.actionButton.setText(self._action_text[Config.ACTION_DENY_IDX])
        self.actionButton.setIcon(self._action_icon[Config.ACTION_DENY_IDX])
        if self._default_action != Config.ACTION_ALLOW_IDX:
            self.actionButton.setText(self._action_text[self._default_action])
            self.actionButton.setIcon(self._action_icon[self._default_action])
        self.actionButton.clicked.connect(self._on_deny_btn_clicked)

    def _configure_plugins(self):
        """configure the plugins that apply to this dialog.
        When configuring the plugins on a particular view, they'll add,
        change or extend the existing functionality.
        """
        for conf in self._action_list:
            action = self._action_list[conf]
            for name in action['actions']:
                try:
                    action['actions'][name].configure(self)
                except Exception as e:
                    print("popups._configure_plugins() exception:", name, " you may want to enable this plugin -", e)

    def _pre_popup_plugins(self, con):
        pass

    def _post_popup_plugins(self, conn):
        """Actions performed on the pop-up once the connection details have
        been displayed on the screen.
        """
        if self._action_list == None:
            return

        for conf in self._action_list:
            action = self._action_list[conf]
            for name in action['actions']:
                try:
                    action['actions'][name].run(self, (conn,))
                except Exception as e:
                    print("popups._post_popup_plugins() exception:", name, "-", e)

    def eventFilter(self, obj, event):
        if event.type() == QEvent.MouseButtonPress:
            self._stop_countdown()
            return True
        return False

    def showEvent(self, event):
        super(PromptDialog, self).showEvent(event)
        self.activateWindow()
        self.adjust_size()
        self.move_popup()

    def adjust_size(self):
        if self._width is None or self._height is None:
            self._width = self.width()
            self._height = self.height()

        self.resize(QtCore.QSize(self._width, self._height))

    def move_popup(self):
        popup_pos = self._cfg.getInt(self._cfg.DEFAULT_POPUP_POSITION)
        point = QtWidgets.QDesktopWidget().availableGeometry()
        if popup_pos == self._cfg.POPUP_TOP_RIGHT:
            self.move(point.topRight())
        elif popup_pos == self._cfg.POPUP_TOP_LEFT:
            self.move(point.topLeft())
        elif popup_pos == self._cfg.POPUP_BOTTOM_RIGHT:
            self.move(point.bottomRight())
        elif popup_pos == self._cfg.POPUP_BOTTOM_LEFT:
            self.move(point.bottomLeft())

    def _stop_countdown(self):
        action_idx = self._cfg.getInt(self._cfg.DEFAULT_ACTION_KEY)
        if action_idx == Config.ACTION_ALLOW_IDX:
            self.allowButton.setText(self._allow_text)
            self.allowButton.setIcon(self.allowIcon)
        else:
            self.actionButton.setText(self._action_text[action_idx])
            self.actionButton.setIcon(self._action_icon[action_idx])
        if self._tick_thread != None:
            self._tick_thread.stop = True

    def _check_advanced_toggled(self, state):
        self.checkDstIP.setVisible(state)
        self.whatIPCombo.setVisible(state)
        self.destIPLabel.setVisible(not state)
        self.checkDstPort.setVisible(state == True and (self._con != None and self._con.dst_port != 0))
        self.checkUserID.setVisible(state)
        self.checkSum.setVisible(self._con.process_checksums[Config.OPERAND_PROCESS_HASH_MD5] != "" and state)
        self.checksumLabel_2.setVisible(self._con.process_checksums[Config.OPERAND_PROCESS_HASH_MD5] != "" and state)
        self.checksumLabel.setVisible(self._con.process_checksums[Config.OPERAND_PROCESS_HASH_MD5] != "" and state)
        self.stackedWidget.setCurrentIndex(_constants.PAGE_MAIN)

        self._ischeckAdvanceded = state
        self.adjust_size()
        self.move_popup()

    def _button_clicked(self):
        self._stop_countdown()

    def _cb_warninglbl_clicked(self, link):
        self._stop_countdown()
        if link == "#warning-checksum":
            self.stackedWidget.setCurrentIndex(_constants.PAGE_CHECKSUMS)

    def _cb_cmdinfo_clicked(self):
        self.stackedWidget.setCurrentIndex(_constants.PAGE_DETAILS)
        self._stop_countdown()

    def _cb_update_rule_clicked(self, updateAll=False):
        self.labelChecksumStatus.setStyleSheet('')
        curRule = self.comboChecksumRule.currentText()
        if curRule == "":
            return

        for idx in range(0, self.comboChecksumRule.count()):
            curRule = self.comboChecksumRule.itemText(idx)

            rule, error = _checksums.update_rule(self._peer, self._rules, curRule, self._con)
            if rule == None:
                self.labelChecksumStatus.setStyleSheet('color: red')
                self.labelChecksumStatus.setText("✘ " + error)
                return

            self._nodes.send_notification(
                self._peer,
                ui_pb2.Notification(
                    id=int(str(time.time()).replace(".", "")),
                    type=ui_pb2.CHANGE_RULE,
                    data="",
                    rules=[rule]
                )
            )
            self.labelChecksumStatus.setStyleSheet('color: green')
            self.labelChecksumStatus.setText("✔" + QC.translate("popups", "Rule updated."))

            if not updateAll:
                break

    def _cb_cmdback_clicked(self):
        self.stackedWidget.setCurrentIndex(_constants.PAGE_MAIN)
        self._stop_countdown()

    def promptUser(self, connection, is_local, peer):
        # one at a time
        with self._lock:
            # reset state
            if self._tick_thread != None and self._tick_thread.is_alive():
                self._tick_thread.join()
            self._cfg.reload()
            self._tick = int(self._cfg.getSettings(self._cfg.DEFAULT_TIMEOUT_KEY)) if self._cfg.hasKey(self._cfg.DEFAULT_TIMEOUT_KEY) else _constants.DEFAULT_TIMEOUT
            self._tick_thread = threading.Thread(target=self._timeout_worker)
            self._tick_thread.stop = self._ischeckAdvanceded
            self._timeout_triggered = False
            self._rule = None
            self._local = is_local
            self._con = connection

            # XXX: workaround for protobufs that don't report the address of
            # the node. In this case the addr is "unix:/local"
            proto, addr = self._nodes.get_addr(peer)
            self._peer = proto
            if addr != None:
                self._peer = proto+":"+addr

            self._done.clear()
            # trigger and show dialog
            self._prompt_trigger.emit()
            # start timeout thread
            self._tick_thread.start()
            # wait for user choice or timeout
            self._done.wait()

            return self._rule, self._timeout_triggered

    def _timeout_worker(self):
        if self._tick == 0:
            self._timeout_trigger.emit()
            return

        while self._tick > 0 and self._done.is_set() is False:
            t = threading.currentThread()
            # stop only stops the coundtdown, not the thread itself.
            if getattr(t, "stop", True):
                self._tick = int(self._cfg.getSettings(self._cfg.DEFAULT_TIMEOUT_KEY))
                time.sleep(1)
                continue

            self._tick -= 1
            self._tick_trigger.emit()
            time.sleep(1)

        if not self._done.is_set():
            self._timeout_trigger.emit()

    @QtCore.pyqtSlot()
    def on_connection_prompt_triggered(self):
        self.stackedWidget.setCurrentIndex(_constants.PAGE_MAIN)
        self._render_connection(self._con)
        if self._tick > 0:
            self.show()
        # render details after displaying the pop-up.

        self._display_checksums_warning(self._peer, self._con)
        _details.render(self._peer, self.connDetails, self._con)

    @QtCore.pyqtSlot()
    def on_tick_triggered(self):
        self._set_cmd_action_text()

    @QtCore.pyqtSlot()
    def on_timeout_triggered(self):
        self._timeout_triggered = True
        self._send_rule()

    def _hide_widget(self, widget, hide):
        widget.setVisible(not hide)

    def _set_cmd_action_text(self):
        action_idx = self._cfg.getInt(self._cfg.DEFAULT_ACTION_KEY)
        if action_idx == Config.ACTION_ALLOW_IDX:
            self.allowButton.setText("{0} ({1})".format(self._allow_text, self._tick))
            self.allowButton.setIcon(self.allowIcon)
            self.actionButton.setText(self._action_text[Config.ACTION_DENY_IDX])
        else:
            self.allowButton.setText(self._allow_text)
            self.actionButton.setText("{0} ({1})".format(self._action_text[action_idx], self._tick))
            self.actionButton.setIcon(self._action_icon[action_idx])

    def _display_checksums_warning(self, peer, con):
        self.messageLabel.setStyleSheet('')
        self.labelChecksumStatus.setText('')
        is_valid = True
        checksums = con.process_checksums
        expected_list = []

        records = self._rules.get_by_field(peer, "operator_data", con.process_path)

        if records != None and records.first():
            rules_names = []
            while True:
                if not records.next():
                    break
                rule = Rule.new_from_records(records)

                if not rule.enabled:
                    continue
                rules_names.append(rule.name)
                validates, expected = _checksums.verify(checksums, rule)
                if not validates:
                    expected_list.append(expected)
                is_valid &= validates

            if is_valid:
                return ""

            self.messageLabel.setStyleSheet('color: red')
            self.messageLabel.setText(
                QC.translate("popups", "WARNING, bad checksum (<a href='#warning-checksum'>More info</a>)"
                                )
            )
            self.labelChecksumNote.setText(
                QC.translate(
                    "popups",
                    "<font color=\"red\">WARNING, checksums differ for at least one rule.</font><br><br>Current process ({0}):<br>{1}<br><br>Expected from the rule:<br>{2}"
                    .format(
                        con.process_id,
                        checksums[Config.OPERAND_PROCESS_HASH_MD5],
                        expected_list
                    ))
            )

            self.comboChecksumRule.clear()
            self.comboChecksumRule.addItems(rules_names)

            return "<b>WARNING</b><br>bad md5<br>This process:{0}<br>Expected from rule: {1}<br><br>".format(
                checksums[Config.OPERAND_PROCESS_HASH_MD5],
                expected
            )

        return ""

    def _render_connection(self, con):
        app_name, app_icon, description, _ = self._apps_parser.get_info_by_path(con.process_path, "terminal")
        app_args = " ".join(con.process_args)
        _utils.set_app_description(self.appDescriptionLabel, description)
        _utils.set_app_path(self.appPathLabel, app_name, app_args, con)
        _utils.set_app_args(self.argsLabel, app_name, app_args)

        self.checksumLabel.setText(con.process_checksums[Config.OPERAND_PROCESS_HASH_MD5])
        self.checkSum.setChecked(False)

        if app_name == "":
            self.appPathLabel.setVisible(False)
            self.argsLabel.setVisible(False)
            self.argsLabel.setText("")
            app_name = QC.translate("popups", "Unknown process %s" % con.process_path)
            #with self._lock:
            self.appNameLabel.setText(QC.translate("popups", "Outgoing connection"))
        else:
            _utils.set_elide_text(self.appNameLabel, "%s" % app_name, max_size=42)
            self.appNameLabel.setToolTip(app_name)

        #if len(self._con.process_args) == 0 or self._con.process_args[0] == "":

        self.cwdLabel.setToolTip("%s %s" % (QC.translate("popups", "Process launched from:"), con.process_cwd))
        _utils.set_elide_text(self.cwdLabel, con.process_cwd, max_size=32)

        pixmap = Icons.get_by_appname(app_icon)
        self.iconLabel.setPixmap(pixmap)

        message = _utils.get_popup_message(self._local, self._peer, app_name, con)

        self.messageLabel.setText(message)
        self.messageLabel.setToolTip(message)

        self.sourceIPLabel.setText(con.src_ip)
        self.destIPLabel.setText(con.dst_ip)
        if con.dst_port == 0:
            self.destPortLabel.setText("")
        else:
            self.destPortLabel.setText(str(con.dst_port))
        self._hide_widget(self.destPortLabel, con.dst_port == 0)
        self._hide_widget(self.checkSum, con.process_checksums[Config.OPERAND_PROCESS_HASH_MD5] == "" or not self._ischeckAdvanceded)
        self._hide_widget(self.checksumLabel, con.process_checksums[Config.OPERAND_PROCESS_HASH_MD5] == "" or not self._ischeckAdvanceded)
        self._hide_widget(self.checksumLabel_2, con.process_checksums[Config.OPERAND_PROCESS_HASH_MD5] == "" or not self._ischeckAdvanceded)
        self._hide_widget(self.destPortLabel_1, con.dst_port == 0)
        self._hide_widget(self.checkDstPort, con.dst_port == 0 or not self._ischeckAdvanceded)

        if self._local:
            try:
                uid = "{0} ({1})".format(con.user_id, pwd.getpwuid(con.user_id).pw_name)
            except:
                uid = ""
        else:
            uid = "{0}".format(con.user_id)

        self.uidLabel.setText(uid)

        self.whatCombo.clear()
        self.whatIPCombo.clear()

        self._add_fixed_options_to_combo(self.whatCombo, con, uid)
        if con.process_path.startswith(_constants.APPIMAGE_PREFIX):
            self._add_appimage_pattern_to_combo(self.whatCombo, con)
        self._add_dst_networks_to_combo(self.whatCombo, con.dst_ip)

        if con.dst_host != "" and con.dst_host != con.dst_ip:
            self._add_dsthost_to_combo(con.dst_host)

        self._add_ip_regexp_to_combo(self.whatCombo, self.whatIPCombo, con)
        self._add_dst_networks_to_combo(self.whatIPCombo, con.dst_ip)

        self._default_action = self._cfg.getInt(self._cfg.DEFAULT_ACTION_KEY)
        _utils.set_default_duration(self._cfg, self.durationCombo)

        _utils.set_default_target(self.whatCombo, con, self._cfg, app_name, app_args)

        self.checkDstIP.setChecked(self._cfg.getBool(self._cfg.DEFAULT_POPUP_ADVANCED_DSTIP))
        self.checkDstPort.setChecked(self._cfg.getBool(self._cfg.DEFAULT_POPUP_ADVANCED_DSTPORT))
        self.checkUserID.setChecked(self._cfg.getBool(self._cfg.DEFAULT_POPUP_ADVANCED_UID))
        self.checkSum.setChecked(self._cfg.getBool(self._cfg.DEFAULT_POPUP_ADVANCED_CHECKSUM))
        if self._cfg.getBool(self._cfg.DEFAULT_POPUP_ADVANCED):
            self.checkAdvanced.toggle()

        self._set_cmd_action_text()
        self.checkAdvanced.setFocus()

        self.setFixedSize(self.size())

        self._post_popup_plugins(con)

    # https://gis.stackexchange.com/questions/86398/how-to-disable-the-escape-key-for-a-dialog
    def keyPressEvent(self, event):
        if not event.key() == QtCore.Qt.Key_Escape:
            super(PromptDialog, self).keyPressEvent(event)

    # prevent a click on the window's x
    # from quitting the whole application
    def closeEvent(self, e):
        self._send_rule()
        e.ignore()

    def close(self):
        self._stop_countdown()
        self._done.set()
        self.hide()

    def _add_fixed_options_to_combo(self, combo, con, uid):
        # the order of these combobox entries must match those in the preferences dialog
        # prefs -> UI -> Default target
        combo.addItem(QC.translate("popups", "from this executable"), _constants.FIELD_PROC_PATH)
        if int(con.process_id) < 0:
            combo.model().item(0).setEnabled(False)

        combo.addItem(QC.translate("popups", "from this command line"), _constants.FIELD_PROC_ARGS)

        combo.addItem(QC.translate("popups", "to port {0}").format(con.dst_port), _constants.FIELD_DST_PORT)
        combo.addItem(QC.translate("popups", "to {0}").format(con.dst_ip), _constants.FIELD_DST_IP)

        combo.addItem(QC.translate("popups", "from user {0}").format(uid), _constants.FIELD_USER_ID)
        if int(con.user_id) < 0:
            combo.model().item(4).setEnabled(False)

        combo.addItem(QC.translate("popups", "from this PID"), _constants.FIELD_PROC_ID)

    def _add_ip_regexp_to_combo(self, combo, IPcombo, con):
        IPcombo.addItem(QC.translate("popups", "to {0}").format(con.dst_ip), _constants.FIELD_DST_IP)

        parts = con.dst_ip.split('.')
        nparts = len(parts)
        for i in range(1, nparts):
            combo.addItem(QC.translate("popups", "to {0}.*").format('.'.join(parts[:i])), _constants.FIELD_REGEX_IP)
            IPcombo.addItem(QC.translate("popups", "to {0}.*").format( '.'.join(parts[:i])), _constants.FIELD_REGEX_IP)

    def _add_appimage_pattern_to_combo(self, combo, con):
        """appimages' absolute path usually starts with /tmp/.mount_<
        """
        appimage_bin = os.path.basename(con.process_path)
        appimage_path = os.path.dirname(con.process_path)
        appimage_path = appimage_path[0:len(_constants.APPIMAGE_PREFIX)+6]
        combo.addItem(
            QC.translate("popups", "from {0}*/{1}").format(appimage_path, appimage_bin),
            _constants.FIELD_APPIMAGE
        )

    def _add_dst_networks_to_combo(self, combo, dst_ip):
        alias = NetworkAliases.get_alias(dst_ip)
        if alias:
            combo.addItem(QC.translate("popups", f"to {alias}"), _constants.FIELD_DST_NETWORK)
        if type(ipaddress.ip_address(dst_ip)) == ipaddress.IPv4Address:
            combo.addItem(QC.translate("popups", "to {0}").format(ipaddress.ip_network(dst_ip + "/24", strict=False)),  _constants.FIELD_DST_NETWORK)
            combo.addItem(QC.translate("popups", "to {0}").format(ipaddress.ip_network(dst_ip + "/16", strict=False)),  _constants.FIELD_DST_NETWORK)
            combo.addItem(QC.translate("popups", "to {0}").format(ipaddress.ip_network(dst_ip + "/8", strict=False)),   _constants.FIELD_DST_NETWORK)
        else:
            combo.addItem(QC.translate("popups", "to {0}").format(ipaddress.ip_network(dst_ip + "/64", strict=False)),  _constants.FIELD_DST_NETWORK)
            combo.addItem(QC.translate("popups", "to {0}").format(ipaddress.ip_network(dst_ip + "/128", strict=False)), _constants.FIELD_DST_NETWORK)

    def _add_dsthost_to_combo(self, dst_host):
        self.whatCombo.addItem("%s" % dst_host, _constants.FIELD_DST_HOST)
        self.whatIPCombo.addItem("%s" % dst_host, _constants.FIELD_DST_HOST)

        parts = dst_host.split('.')[1:]
        nparts = len(parts)
        for i in range(0, nparts - 1):
            self.whatCombo.addItem(QC.translate("popups", "to *.{0}").format('.'.join(parts[i:])), _constants.FIELD_REGEX_HOST)
            self.whatIPCombo.addItem(QC.translate("popups", "to *.{0}").format('.'.join(parts[i:])), _constants.FIELD_REGEX_HOST)

    def _on_action_clicked(self, action):
        self._default_action = action
        self._send_rule()

    def _on_deny_btn_clicked(self, action):
        self._default_action = self._cfg.getInt(self._cfg.DEFAULT_ACTION_KEY)
        if self._default_action == Config.ACTION_ALLOW_IDX:
            self._default_action = Config.ACTION_DENY_IDX
        self._send_rule()

    def _is_list_rule(self):
        return self.checkUserID.isChecked() or \
            self.checkDstPort.isChecked() or \
            self.checkDstIP.isChecked() or \
            self.checkSum.isChecked()

    def _send_rule(self):
        try:
            self._cfg.setSettings("promptDialog/geometry", self.saveGeometry())
            self._rule = ui_pb2.Rule(name="user.choice")
            self._rule.created = int(datetime.now().timestamp())
            self._rule.enabled = True
            self._rule.duration = _utils.get_duration(self.durationCombo.currentIndex())

            self._rule.action = Config.ACTION_ALLOW
            if self._default_action == Config.ACTION_DENY_IDX:
                self._rule.action = Config.ACTION_DENY
            elif self._default_action == Config.ACTION_REJECT_IDX:
                self._rule.action = Config.ACTION_REJECT

            what_idx = self.whatCombo.currentIndex()
            self._rule.operator.type, self._rule.operator.operand, self._rule.operator.data = _utils.get_combo_operator(
                self.whatCombo.itemData(what_idx),
                self.whatCombo.currentText(),
                self._con)
            if self._rule.operator.data == "":
                print("popups: Invalid rule, discarding: ", self._rule)
                self._rule = None
                return

            rule_temp_name = _utils.get_rule_name(self._rule, self._is_list_rule())
            self._rule.name = rule_temp_name

            # TODO: move to a method
            data=[]

            alias_selected = False

            if self.whatCombo.itemData(what_idx) == _constants.FIELD_DST_NETWORK:
                alias = NetworkAliases.get_alias(self._con.dst_ip)
                if alias:
                    _type, _operand, _data = Config.RULE_TYPE_SIMPLE, Config.OPERAND_PROCESS_PATH, self._con.process_path
                    data.append({"type": _type, "operand": _operand, "data": _data})
                    rule_temp_name = slugify(f"{rule_temp_name} {os.path.basename(self._con.process_path)}")
                    alias_selected = True

            if self.checkDstIP.isChecked() and self.whatCombo.itemData(what_idx) != _constants.FIELD_DST_IP:
                _type, _operand, _data = _utils.get_combo_operator(
                    self.whatIPCombo.itemData(self.whatIPCombo.currentIndex()),
                    self.whatIPCombo.currentText(),
                    self._con)
                data.append({"type": _type, "operand": _operand, "data": _data})
                rule_temp_name = slugify("%s %s" % (rule_temp_name, _data))

            if self.checkDstPort.isChecked() and self.whatCombo.itemData(what_idx) != _constants.FIELD_DST_PORT:
                data.append({"type": Config.RULE_TYPE_SIMPLE, "operand": Config.OPERAND_DEST_PORT, "data": str(self._con.dst_port)})
                rule_temp_name = slugify("%s %s" % (rule_temp_name, str(self._con.dst_port)))

            if self.checkUserID.isChecked() and self.whatCombo.itemData(what_idx) != _constants.FIELD_USER_ID:
                data.append({"type": Config.RULE_TYPE_SIMPLE, "operand": Config.OPERAND_USER_ID, "data": str(self._con.user_id)})
                rule_temp_name = slugify("%s %s" % (rule_temp_name, str(self._con.user_id)))

            if self.checkSum.isChecked() and self.checksumLabel.text() != "":
                _type, _operand, _data = Config.RULE_TYPE_SIMPLE, Config.OPERAND_PROCESS_HASH_MD5, self.checksumLabel.text()
                data.append({"type": _type, "operand": _operand, "data": _data})
                rule_temp_name = slugify("%s %s" % (rule_temp_name, _operand))

            is_list_rule = self._is_list_rule()

            # If the user has selected to filter by cmdline, but the launched
            # command path is not absolute or the first component contains
            # "/proc/" (/proc/self/fd.., /proc/1234/fd...), we can't trust it.
            # In these cases, also filter by the absolute path to the binary.
            if self._rule.operator.operand == Config.OPERAND_PROCESS_COMMAND:
                proc_args = " ".join(self._con.process_args)
                proc_args = proc_args.split(" ")
                if os.path.isabs(proc_args[0]) == False or proc_args[0].startswith("/proc"):
                    is_list_rule = True
                    data.append({"type": Config.RULE_TYPE_SIMPLE, "operand": Config.OPERAND_PROCESS_PATH, "data": str(self._con.process_path)})

            if is_list_rule or alias_selected:
                data.append({
                    "type": self._rule.operator.type,
                    "operand": self._rule.operator.operand,
                    "data": self._rule.operator.data
                })
                # We need to send back the operator list to the AskRule() call
                # as json string, in order to add it to the DB.
                self._rule.operator.data = json.dumps(data)
                self._rule.operator.type = Config.RULE_TYPE_LIST
                self._rule.operator.operand = Config.RULE_TYPE_LIST
                for op in data:
                    self._rule.operator.list.extend([
                        ui_pb2.Operator(
                            type=op['type'],
                            operand=op['operand'],
                            sensitive=False if op.get('sensitive') == None else op['sensitive'],
                            data="" if op.get('data') == None else op['data']
                        )
                    ])

            exists = self._rules.exists(self._rule, self._peer)
            if not exists:
                self._rule.name = self._rules.new_unique_name(rule_temp_name, self._peer, "")

            self.hide()
            if self._ischeckAdvanceded:
                self.checkAdvanced.toggle()
            self._ischeckAdvanceded = False

        except Exception as e:
            print("[pop-up] exception creating a rule:", e)
        finally:
            # signal that the user took a decision and
            # a new rule is available
            self._done.set()
            self.hide()
